from dataclasses import dataclass, field
from enum import IntEnum
from typing import Annotated, Any

from pydantic import Discriminator, Tag

from ._base_node import LayoutNode
from .sources import DataSource, LiteralSource, Source, SourceRef, TransformOutputRoleRef
from .sources.aggregation import AggregationSource, ScopedEvalExpression, SelectRef
from .sources.arithmetic import ArithmeticSource, ScopedEval
from .sources.column import ColumnSource
from .sources.group import GroupSource
from .sources.proto import ProtoSourceRef


class ExpressionVersion(IntEnum):
    VERSION_1 = 1
    VERSION_2 = 2


class _AnyValueHelper(LayoutNode):
    DefaultValueOverridesAncestors: bool


class AnyValue(LayoutNode):
    AnyValue: _AnyValueHelper


class QueryConditionType(IntEnum):
    """Names defined by myself, but based on query outputs from the query tester."""

    STANDARD = 0
    TOP_N = 2
    MEASURE = 3


class ComparisonKind(IntEnum):
    IS_EQUAL = 0
    IS_GREATER_THAN = 1
    IS_GREATER_THAN_OR_EQUAL_TO = 2
    IS_LESS_THAN = 3
    IS_LESS_THAN_OR_EQUAL_TO = 4

    def get_operator(self) -> str:
        OPERATOR_MAPPING = {  # noqa: N806
            ComparisonKind.IS_EQUAL: "=",
            ComparisonKind.IS_GREATER_THAN: ">",
            ComparisonKind.IS_GREATER_THAN_OR_EQUAL_TO: ">=",
            ComparisonKind.IS_LESS_THAN: "<",
            ComparisonKind.IS_LESS_THAN_OR_EQUAL_TO: "<=",
        }
        if self not in OPERATOR_MAPPING:
            msg = f"No operator is defined for: {self}"
            raise ValueError(msg)
        return OPERATOR_MAPPING[self]


class ComparisonHelper(LayoutNode):
    Left: DataSource
    Right: LiteralSource


@dataclass
class Expression:
    template: str
    source: str
    data: dict[str, str] = field(default_factory=dict)
    expr_type: str = ""

    def to_text(self) -> str:
        if self.data:
            return self.template.format(**self.data)
        return self.template


class ContainsCondition(LayoutNode):
    Contains: ComparisonHelper

    def natural_language(self) -> str:
        """Returns a natural language representation of the condition."""
        left = natural_language_source(self.Contains.Left)
        right = natural_language_source(self.Contains.Right)
        return f"{left} CONTAINS {right}"

    def get_sources(self) -> list[DataSource]:
        """Returns the sources used in the condition."""
        if isinstance(self.Contains.Right, LiteralSource):
            return [self.Contains.Left]
        return [self.Contains.Left, self.Contains.Right]


class InExpressionHelper(LayoutNode):
    Expressions: list[DataSource]
    Values: list[list[LiteralSource]]

    def vals(self) -> list[str]:
        return [str(y.value()) for x in self.Values for y in x]

    def __repr__(self) -> str:
        source = self.Expressions[0].__repr__()
        return f"In({source}, {', '.join(self.vals())})"

    def get_sources(self) -> list[DataSource]:
        return self.Expressions


class InTopNExpressionHelper(LayoutNode):
    """Internal representation of the Top N option."""

    Expressions: list[DataSource]
    Table: SourceRef

    def get_sources(self) -> list[DataSource]:
        return self.Expressions


def get_in_union_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "Table" in v:
            return "InTopNExpressionHelper"
        if "Values" in v:
            return "InExpressionHelper"
        raise TypeError(v)
    return v.__class__.__name__


InUnion = Annotated[
    Annotated[InExpressionHelper, Tag("InExpressionHelper")]
    | Annotated[InTopNExpressionHelper, Tag("InTopNExpressionHelper")],
    Discriminator(get_in_union_type),
]


def natural_language_source(d: Source | SourceRef | ScopedEvalExpression) -> str:
    if isinstance(d, ColumnSource):
        return d.Column.Property
    breakpoint()
    msg = f"Unsupported data source type: {d.__class__.__name__}"
    raise TypeError(msg)


class InCondition(LayoutNode):
    """In is how "is" and "is not" are internally represented."""

    In: InUnion

    def __repr__(self) -> str:
        return self.In.__repr__()

    def natural_language(self) -> str:
        expr = natural_language_source(self.In.Expressions[0])
        if isinstance(self.In, InTopNExpressionHelper):
            table = natural_language_source(self.In.Table)
            return f"{expr} IN TOP N {table}"
        return f"{expr} IN ({', '.join(str(x[0].value()) for x in self.In.Values)})"

    def get_sources(self) -> list[DataSource]:
        return self.In.get_sources()


class TimeUnit(IntEnum):
    SECOND = 1
    MINUTE = 2
    HOUR = 3
    DAY = 4
    WEEK = 5
    MONTH = 6
    QUARTER = 7
    YEAR = 8


class _NowHelper(LayoutNode):
    Now: dict[str, str]  # actually an empty string


def get_date_span_union_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "Literal" in v:
            return "LiteralSource"
        if "Now" in v:
            return "_NowHelper"
        raise TypeError(v)
    return v.__class__.__name__


DateSpanUnion = Annotated[
    Annotated[LiteralSource, Tag("LiteralSource")] | Annotated[_NowHelper, Tag("_NowHelper")],
    Discriminator(get_date_span_union_type),
]


class _DateSpanHelper(LayoutNode):
    Expression: DateSpanUnion
    TimeUnit: TimeUnit


class DateSpan(LayoutNode):
    DateSpan: _DateSpanHelper


class RangePercentHelper(LayoutNode):
    Min: ScopedEval
    Max: ScopedEval
    Percent: float


class RangePercent(LayoutNode):
    RangePercent: RangePercentHelper


def get_comparison_right_union_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "Literal" in v:
            return "LiteralSource"
        if "AnyValue" in v:
            return "AnyValue"
        if "DateSpan" in v:
            return "DateSpan"
        if "RangePercent" in v:
            return "RangePercent"
        raise TypeError(v)
    return v.__class__.__name__


ComparisonRightUnion = Annotated[
    Annotated[LiteralSource, Tag("LiteralSource")]
    | Annotated[AnyValue, Tag("AnyValue")]
    | Annotated[DateSpan, Tag("DateSpan")]
    | Annotated[RangePercent, Tag("RangePercent")],
    Discriminator(get_comparison_right_union_type),
]


class ComparisonConditionHelper(LayoutNode):
    ComparisonKind: ComparisonKind
    Left: ScopedEvalExpression
    Right: ComparisonRightUnion


class ComparisonCondition(LayoutNode):
    Comparison: ComparisonConditionHelper

    def natural_language(self) -> str:
        """Returns a natural language representation of the condition."""
        left = natural_language_source(self.Comparison.Left)
        right = (
            self.Comparison.Right.value()
            if isinstance(self.Comparison.Right, LiteralSource)
            else str(self.Comparison.Right)
        )
        operator = self.Comparison.ComparisonKind.get_operator()
        return f"{left} {operator} {right}"

    def get_sources(self) -> list[DataSource]:
        left = self.Comparison.Left
        if isinstance(left, AggregationSource):
            return left.get_sources()
        if isinstance(left, SelectRef):
            return []
        return [left]


class NotConditionHelper(LayoutNode):
    Expression: "ConditionType"


class NotCondition(LayoutNode):
    Not: NotConditionHelper

    def __repr__(self) -> str:
        return f"Not({self.Not.Expression.__repr__()})"

    def natural_language(self) -> str:
        """Returns a natural language representation of the condition."""
        return f"NOT {self.Not.Expression.natural_language()}"

    def get_sources(self) -> list[DataSource]:
        return self.Not.Expression.get_sources()


class ExistsConditionHelper(LayoutNode):
    Expression: Source  # cannot be DataSource, might only be a ProtoSourceRef?


class ExistsCondition(LayoutNode):
    Exists: ExistsConditionHelper

    def natural_language(self) -> str:
        """Returns a natural language representation of the condition."""
        return f"Exists({natural_language_source(self.Exists.Expression)})"

    def get_sources(self) -> list[DataSource]:
        expr = self.Exists.Expression
        if isinstance(expr, (AggregationSource, ArithmeticSource, GroupSource)):
            return expr.get_sources()
        if isinstance(expr, (ProtoSourceRef, SelectRef, LiteralSource, TransformOutputRoleRef)):
            return []
        return [
            expr,
        ]


class CompositeConditionHelper(LayoutNode):
    Left: "ConditionType"
    Right: "ConditionType"


class AndCondition(LayoutNode):
    And: CompositeConditionHelper

    def natural_language(self) -> str:
        """Returns a natural language representation of the condition."""
        return f"({self.And.Left.natural_language()} AND {self.And.Right.natural_language()})"

    def get_sources(self) -> list[DataSource]:
        return [*self.And.Left.get_sources(), *self.And.Right.get_sources()]


class OrCondition(LayoutNode):
    Or: CompositeConditionHelper

    def natural_language(self) -> str:
        """Returns a natural language representation of the condition."""
        return f"({self.Or.Left.natural_language()} OR {self.Or.Right.natural_language()})"

    def get_sources(self) -> list[DataSource]:
        return [
            *self.Or.Left.get_sources(),
            *self.Or.Right.get_sources(),
        ]


def get_type(v: object | dict[str, Any]) -> str:  # noqa: PLR0911
    if isinstance(v, dict):
        if "And" in v:
            return "AndCondition"
        if "Or" in v:
            return "OrCondition"
        if "Left" in v:
            return "NonCompositeConditions"
        if "In" in v:
            return "InCondition"
        if "Not" in v:
            return "NotCondition"
        if "Contains" in v:
            return "ContainsCondition"
        if "Comparison" in v:
            return "ComparisonCondition"
        if "Exists" in v:
            return "ExistsCondition"
        raise ValueError
    return v.__class__.__name__


ConditionType = Annotated[
    Annotated[AndCondition, Tag("AndCondition")]
    | Annotated[OrCondition, Tag("OrCondition")]
    | Annotated[InCondition, Tag("InCondition")]
    | Annotated[NotCondition, Tag("NotCondition")]
    | Annotated[ContainsCondition, Tag("ContainsCondition")]
    | Annotated[ComparisonCondition, Tag("ComparisonCondition")]
    | Annotated[ExistsCondition, Tag("ExistsCondition")],
    Discriminator(get_type),
]


class Condition(LayoutNode):
    Condition: ConditionType
    Target: list[Source] | None = None

    def __repr__(self) -> str:
        return f"Condition({self.Condition.__repr__()})"

    def natural_language(self) -> str:
        """Returns a natural language representation of the condition."""
        return self.Condition.natural_language()

    def get_sources(self) -> list[DataSource]:
        """Returns the sources used in the condition.

        Note: The left source must come first, since the
            order is used by PowerBI and this library
            to identify the default display name of filters
        """
        return self.Condition.get_sources()



from enum import IntEnum


class AuditLoginColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    SEVERITY = 22  # Severity level of an exception.
    SUCCESS = 23  # 1 = success. 0 = failure (for example, a 1 means success of a permissions check and a 0 means a failure of that check).
    ERROR = 24  # Error number of a given event.
    CONNECTIONID = 25  # Unique connection ID.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTHOSTNAME = 35  # Name of the computer on which the client is running. This data column is populated if the host name is provided by the client.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    APPLICATIONNAME = 37  # Name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47


class AuditLogoutColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    ENDTIME = 4  # Time at which the event ended. This column is not populated for starting event classes, such as SQL:BatchStarting or SP:Starting. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    DURATION = 5  # Amount of time (in milliseconds) taken by the event.
    CPUTIME = 6  # Amount of CPU time (in milliseconds) used by the event.
    SUCCESS = 23  # 1 = success. 0 = failure (for example, a 1 means success of a permissions check and a 0 means a failure of that check).
    CONNECTIONID = 25  # Unique connection ID.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTHOSTNAME = 35  # Name of the computer on which the client is running. This data column is populated if the host name is provided by the client.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    APPLICATIONNAME = 37  # Name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47


class AuditServerStartsAndStopsColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    EVENTSUBCLASS = 1  # Event Subclass provides additional information about each event class: 1: Instance Shutdown 2: Instance Started 3: Instance Paused 4: Instance Continued
    CURRENTTIME = 2  # Time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    SEVERITY = 22  # Severity level of an exception.
    SUCCESS = 23  # 1 = success. 0 = failure (for example, a 1 means success of a permissions check and a 0 means a failure of that check).
    ERROR = 24  # Error number of a given event.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47


class AuditObjectPermissionEventColumns(IntEnum):
    OBJECTID = 11  # Object ID (note this is a string).
    OBJECTTYPE = 12  # Object type.
    OBJECTNAME = 13  # Object name.
    OBJECTPATH = 14  # Object path. A comma-separated list of parents, starting with the object's parent.
    OBJECTREFERENCE = 15  # Object reference. Encoded as XML for all parents, using tags to describe the object.
    SEVERITY = 22  # Severity level of an exception.
    SUCCESS = 23  # 1 = success. 0 = failure (for example, a 1 means success of a permissions check and a 0 means a failure of that check).
    ERROR = 24  # Error number of a given event.
    CONNECTIONID = 25  # Unique connection ID.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTHOSTNAME = 35  # Name of the computer on which the client is running. This data column is populated if the host name is provided by the client.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    APPLICATIONNAME = 37  # Name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    SESSIONID = 39  # Session GUID.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SPID = 41  # Server process ID. This uniquely identifies a user session. This directly corresponds to the session GUID used by XML/A.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47


class AuditAdminOperationsEventColumns(IntEnum):
    EVENTSUBCLASS = 1  # Event Subclass provides additional information about each event class: 1: Backup 2: Restore 3: Synchronize 4: Detach 5: Attach 6: ImageLoad 7: ImageSave
    SEVERITY = 22  # Severity level of an exception.
    SUCCESS = 23  # 1 = success. 0 = failure (for example, a 1 means success of a permissions check and a 0 means a failure of that check).
    ERROR = 24  # Error number of a given event.
    CONNECTIONID = 25  # Unique connection ID.
    DATABASENAME = 28  # Name of the database in which the statement of the user is running.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTHOSTNAME = 35  # Name of the computer on which the client is running. This data column is populated if the host name is provided by the client.
    CLIENTPROCESSID = 36  # The process ID of the client application.
    APPLICATIONNAME = 37  # Name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    SESSIONID = 39  # Session GUID.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SPID = 41  # Server process ID. This uniquely identifies a user session. This directly corresponds to the session GUID used by XML/A.
    TEXTDATA = 42  # Text data associated with the event.
    SERVERNAME = 43  # Name of the server producing the event.
    ACTIVITYID = 46
    REQUESTID = 47



from typing import TYPE_CHECKING, Literal

from pydantic import PrivateAttr

from pbi_core.lineage import LineageNode
from pbi_core.ssas.model_tables.base import SsasEditableRecord
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.expression import Expression
    from pbi_core.ssas.model_tables.model import Model
    from pbi_core.ssas.model_tables.partition import Partition


class QueryGroup(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/40b3830b-25ee-41a6-87d2-49616028dd13)
    This class represents a group of queries that can be executed together.
    """

    _repr_name_field = "folder"

    description: str | None = None
    folder: str
    model_id: int

    _commands: BaseCommands = PrivateAttr(default_factory=lambda: SsasCommands.query_group)

    def expressions(self) -> set["Expression"]:
        return self.tabular_model.expressions.find_all({"query_group_id": self.id})

    def partitions(self) -> set["Partition"]:
        return self.tabular_model.partitions.find_all({"query_group_id": self.id})

    def model(self) -> "Model":
        return self.tabular_model.model

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(
                self,
                lineage_type,
                [expression.get_lineage(lineage_type) for expression in self.expressions()]
                + [partition.get_lineage(lineage_type) for partition in self.partitions()],
            )
        return LineageNode(self, lineage_type, [self.model().get_lineage(lineage_type)])



import dataclasses

import bs4
import jinja2

BASE_ALTER_TEMPLATE = jinja2.Template(
    """
<Batch Transaction="false" xmlns="http://schemas.microsoft.com/analysisservices/2003/engine">
  <Alter xmlns="http://schemas.microsoft.com/analysisservices/2014/engine">
    <DatabaseID>{{db_name}}</DatabaseID>
    {{entity_def}}
  </Alter>
</Batch>
""".strip(),
)

# note that Transaction = true. I think it's necessary, not very tested tbqh
BASE_REFRESH_TEMPLATE = jinja2.Template(
    """
<Batch Transaction="false" xmlns="http://schemas.microsoft.com/analysisservices/2003/engine">
  <Refresh xmlns="http://schemas.microsoft.com/analysisservices/2014/engine">
	<DatabaseID>{{db_name}}</DatabaseID>
    {{entity_def}}
  </Refresh>
</Batch>
""".strip(),
)
BASE_RENAME_TEMPLATE = jinja2.Template(
    """
<Batch Transaction="false" xmlns="http://schemas.microsoft.com/analysisservices/2003/engine">
  <Alter xmlns="http://schemas.microsoft.com/analysisservices/2014/engine">
    <DatabaseID>{{db_name}}</DatabaseID>
  </Alter>
  <Rename xmlns="http://schemas.microsoft.com/analysisservices/2014/engine">
    <DatabaseID>{{db_name}}</DatabaseID>
    {{entity_def}}
  </Rename>
</Batch>
""".strip(),
)
BASE_DELETE_TEMPLATE = jinja2.Template(
    """
<Batch Transaction="false" xmlns="http://schemas.microsoft.com/analysisservices/2003/engine">
  <Delete xmlns="http://schemas.microsoft.com/analysisservices/2014/engine">
    <DatabaseID>{{db_name}}</DatabaseID>
    {{entity_def}}
  </Delete>
</Batch>
""".strip(),
)
BASE_CREATE_TEMPLATE = jinja2.Template(
    """
<Batch Transaction="false" xmlns="http://schemas.microsoft.com/analysisservices/2003/engine">
  <Create xmlns="http://schemas.microsoft.com/analysisservices/2014/engine">
    <DatabaseID>{{db_name}}</DatabaseID>
    {{entity_def}}
  </Create>
</Batch>
""".strip(),
)
base_commands = {
    "alter": BASE_ALTER_TEMPLATE,
    "create": BASE_CREATE_TEMPLATE,
    "delete": BASE_DELETE_TEMPLATE,
    "refresh": BASE_REFRESH_TEMPLATE,
    "rename": BASE_RENAME_TEMPLATE,
}


@dataclasses.dataclass
class Command:
    entity_template: jinja2.Template
    base_template: jinja2.Template
    field_order: list[str]

    def sort(self, fields: list[tuple[str, str]]) -> list[tuple[str, str]]:
        return sorted(fields, key=lambda k: self.field_order.index(k[0]))


class NoCommands:
    def __init__(self, **kwargs: str) -> None:
        for field_name, template_text in kwargs.items():
            v = Command(
                entity_template=jinja2.Template(template_text),
                base_template=base_commands[field_name],
                field_order=self.get_field_order(template_text),
            )
            self.__setattr__(field_name, v)

    @staticmethod
    def get_field_order(text: str) -> list[str]:
        """Gets the order of the fields for the command, based on the ``xs:sequence`` section of the XML command."""
        tree = bs4.BeautifulSoup(text, "xml")
        row = tree.find("xs:complexType", {"name": "row"})
        assert isinstance(row, bs4.element.Tag)
        ret: list[str] = []
        for e in row.find_all("xs:element"):
            assert isinstance(e, bs4.element.Tag)
            val = e["name"]
            assert isinstance(val, str)
            ret.append(val)
        return ret


class BaseCommands(NoCommands):
    alter: Command
    create: Command
    delete: Command

    def __repr__(self) -> str:
        return "BaseCommands(alter, create, delete)"

    @staticmethod
    def new(data: dict[str, str]) -> "BaseCommands":
        return BaseCommands(
            alter=data["alter.xml"],
            create=data["create.xml"],
            delete=data["delete.xml"],
        )


class RenameCommands(BaseCommands):
    rename: Command

    def __repr__(self) -> str:
        return "RenameCommands(alter, create, delete, rename)"

    @staticmethod
    def new(data: dict[str, str]) -> "RenameCommands":
        return RenameCommands(
            alter=data["alter.xml"],
            create=data["create.xml"],
            delete=data["delete.xml"],
            rename=data["rename.xml"],
        )


class RefreshCommands(RenameCommands):
    refresh: Command

    def __repr__(self) -> str:
        return "RefreshCommands(alter, create, delete, rename, refresh)"

    @staticmethod
    def new(data: dict[str, str]) -> "RefreshCommands":
        return RefreshCommands(
            alter=data["alter.xml"],
            create=data["create.xml"],
            delete=data["delete.xml"],
            rename=data["rename.xml"],
            refresh=data["refresh.xml"],
        )


class ModelCommands(NoCommands):
    alter: Command
    refresh: Command
    rename: Command

    def __repr__(self) -> str:
        return "ModelCommands(alter, refresh, rename)"

    @staticmethod
    def new(data: dict[str, str]) -> "ModelCommands":
        return ModelCommands(
            alter=data["alter.xml"],
            refresh=data["refresh.xml"],
            rename=data["rename.xml"],
        )


Commands = BaseCommands | RenameCommands | RefreshCommands | ModelCommands



from pydantic import Field

from ._base import BaseFileModel

base_val = bool | int | str


class QueryNameMapping(BaseFileModel):
    Key: str
    Value: str


class Metadata(BaseFileModel):
    Version: int
    AutoCreatedRelationships: list[int] = []
    CreatedFrom: str
    CreatedFromRelease: str
    FileDescription: str | None = None
    QueryNameToKeyMapping: list[QueryNameMapping] = Field(alias="_queryNameToKeyMapping", default_factory=list)  # ruff: noqa: N815



from enum import IntEnum


class WlgroupCpuThrottlingColumns(IntEnum):
    EVENTCLASS = 0
    CURRENTTIME = 2
    INTEGERDATA = 10
    OBJECTID = 11
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    APPLICATIONCONTEXT = 52


class WlgroupExceedsMemoryLimitColumns(IntEnum):
    EVENTCLASS = 0
    CURRENTTIME = 2
    STARTTIME = 3
    INTEGERDATA = 10
    OBJECTID = 11
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    APPLICATIONCONTEXT = 52


class WlgroupExceedsProcessingLimitColumns(IntEnum):
    EVENTCLASS = 0
    CURRENTTIME = 2
    INTEGERDATA = 10
    OBJECTID = 11
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    APPLICATIONCONTEXT = 52



from enum import IntEnum


class QueryBeginColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    EVENTSUBCLASS = 1  # Event Subclass provides additional information about each event class. 0: MDXQuery 1: DMXQuery 2: SQLQuery 3: DAXQuery
    CURRENTTIME = 2  # Contains the current time of the event, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Contains the time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    CONNECTIONID = 25  # Contains the unique connection ID associated with the query event.
    DATABASENAME = 28  # Contains the name of the database in which the query is running.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTPROCESSID = 36  # Contains the process ID of the client application.
    APPLICATIONNAME = 37  # Contains the name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    SESSIONID = 39  # Contains the session unique ID of the XMLA request.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SPID = 41  # Contains the server process ID (SPID) that uniquely identifies the user session associated with the query event. The SPID directly corresponds to the session GUID used by XMLA.
    TEXTDATA = 42  # Contains the text data associated with the query event.
    SERVERNAME = 43  # Contains the name of the instance on which the query event occurred.
    REQUESTPARAMETERS = (
        44  # Contains the parameters for parameterized queries and commands associated with the query event.
    )
    REQUESTPROPERTIES = 45  # Contains the properties of the XMLA request.
    ACTIVITYID = 46
    REQUESTID = 47


class QueryEndColumns(IntEnum):
    EVENTCLASS = 0  # Event Class is used to categorize events.
    EVENTSUBCLASS = 1  # Event Subclass provides additional information about each event class. 0: MDXQuery 1: DMXQuery 2: SQLQuery 3: DAXQuery
    CURRENTTIME = 2  # Contains the current time of the event, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    STARTTIME = 3  # Contains the time at which the event started, when available. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    ENDTIME = 4  # Contains the time at which the event ended. This column is not populated for starting event classes, such as SQL:BatchStarting or SP:Starting. For filtering, expected formats are 'YYYY-MM-DD' and 'YYYY-MM-DD HH:MM:SS'.
    DURATION = 5  # Contains the amount of elapsed time (in milliseconds) taken by the event.
    CPUTIME = 6  # Contains the amount of CPU time (in milliseconds) used by the event.
    SEVERITY = 22  # Contains the severity level of an exception associated with the query event. Values are: 0 = Success 1 = Informational 2 = Warning 3 = Error
    SUCCESS = 23  # Contains the success or failure of the query event. Values are: 0 = Failure 1 = Success
    ERROR = 24  # Contains the error number of any error associated with the query event.
    CONNECTIONID = 25  # Contains the unique connection ID associated with the query event.
    DATABASENAME = 28  # Contains the name of the database in which the query is running.
    NTUSERNAME = 32  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service  Principal Name (SPN) (appid@tenantid) - Power BI Service Account  (Power BI Service) - Power BI Service on behalf of a UPN or SPN (Power BI Service (UPN/SPN))
    NTDOMAINNAME = 33  # Contains the domain name associated with the user account that triggered the command event.  - Windows domain name for Windows user accounts - AzureAD for Microsoft Entra accounts - NT AUTHORITY accounts without a Windows domain name, such as the Power BI service
    CLIENTPROCESSID = 36  # Contains the process ID of the client application.
    APPLICATIONNAME = 37  # Contains the name of the client application that created the connection to the server. This column is populated with the values passed by the application rather than the displayed name of the program.
    SESSIONID = 39  # Contains the session unique ID of the XMLA request.
    NTCANONICALUSERNAME = 40  # Contains the user name associated with the command event. Depending on the environment, the user name is in the following form: - Windows user account (DOMAIN\UserName) - User Principal Name (UPN) (username@domain.com) - Service Principal Name (SPN) (appid@tenantid) - Power BI Service Account (Power BI Service)
    SPID = 41  # Contains the server process ID (SPID) that uniquely identifies the user session associated with the query event. The SPID directly corresponds to the session GUID used by XMLA.
    TEXTDATA = 42  # Contains the text data associated with the query event.
    SERVERNAME = 43  # Contains the name of the instance on which the query event occurred.
    ACTIVITYID = 46
    REQUESTID = 47



from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasEditableRecord
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

from .enums import Granularity, PolicyType, RefreshMode


class RefreshPolicy(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/e11ae511-5064-470b-8abc-e2a4dd3999e6)
    This class represents the refresh policy for a partition in a Tabular model.
    """

    incremental_granularity: Granularity
    incremental_periods: int
    incremental_periods_offset: int
    mode: RefreshMode
    policy_type: PolicyType
    polling_expression: str
    rolling_window_granularity: Granularity
    rolling_window_periods: int
    source_expression: str
    table_id: int

    _commands: BaseCommands = PrivateAttr(default_factory=lambda: SsasCommands.refresh_policy)



from enum import IntEnum


class ExecuteSourceQueryColumns(IntEnum):
    EVENTCLASS = 0
    CURRENTTIME = 2
    STARTTIME = 3
    ENDTIME = 4
    DURATION = 5
    CPUTIME = 6
    JOBID = 7
    SESSIONTYPE = 8
    INTEGERDATA = 10
    OBJECTID = 11
    OBJECTTYPE = 12
    OBJECTNAME = 13
    OBJECTPATH = 14
    SEVERITY = 22
    SUCCESS = 23
    ERROR = 24
    CONNECTIONID = 25
    DATABASENAME = 28
    CLIENTPROCESSID = 36
    SESSIONID = 39
    SPID = 41
    TEXTDATA = 42
    SERVERDATA = 43
    ACTIVITYID = 46
    REQUESTID = 47
    ERRORTYPE = 49
    APPLICATIONCONTEXT = 52



from .base_ssas_table import SsasTable
from .enums import RefreshType
from .ssas_tables import SsasEditableRecord, SsasModelRecord, SsasReadonlyRecord, SsasRefreshRecord, SsasRenameRecord

__all__ = [
    "RefreshType",
    "SsasEditableRecord",
    "SsasModelRecord",
    "SsasReadonlyRecord",
    "SsasRefreshRecord",
    "SsasRenameRecord",
    "SsasTable",
]



import datetime
from typing import TYPE_CHECKING

from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasEditableRecord
from pbi_core.ssas.model_tables.enums import DataState, ObjectType
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.measure import Measure
    from pbi_core.ssas.model_tables.table import Table


class DetailRowDefinition(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/7eb1e044-4eed-467d-b10f-ce208798ddb0)

    Note:
        Additional details can be found here: [SQLBI](https://www.sqlbi.com/articles/controlling-drillthrough-in-excel-pivottables-connected-to-power-bi-or-analysis-services/)

    """

    error_message: str
    expression: str
    object_id: int
    object_type: ObjectType
    state: DataState

    modified_time: datetime.datetime

    _commands: BaseCommands = PrivateAttr(default_factory=lambda: SsasCommands.detail_row_defintion)

    @classmethod
    def _db_type_name(cls) -> str:
        return "DetailRowsDefinition"

    def object(self) -> "Table | Measure":
        match self.object_type:
            case ObjectType.MEASURE:
                return self.tabular_model.measures.find(self.object_id)
            case ObjectType.TABLE:
                return self.tabular_model.tables.find(self.object_id)
            case _:
                msg = f"No logic for object type: {self.object_type}"
                raise TypeError(msg)



from typing import Annotated, Any

from pydantic import Discriminator, Tag

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector
from pbi_core.static_files.layout.sources import Source


class TextStyle(LayoutNode):
    color: str | None = None  # TODO: check that it's hex
    fontSize: str | None = None
    fontFamily: str | None = None
    fontStyle: str | None = None  # italic, etc
    fontWeight: str | None = None  # bold, etc
    textDecoration: str | None = None  # underline, etc


class CasePattern(LayoutNode):
    expr: Source


class Case(LayoutNode):
    pattern: CasePattern
    textRuns: list["TextRun"]


class DefaultCaseTextRun(LayoutNode):
    value: str


class DefaultCase(LayoutNode):
    textRuns: list[DefaultCaseTextRun]


class PropertyIdentifier(LayoutNode):
    objectName: str | None = None
    propertyName: str | None = None
    selector: Selector | None = None
    propertyIdentifier: "PropertyIdentifier | None" = None


class TextRunExpression(LayoutNode):
    propertyIdentifier: PropertyIdentifier
    selector: Selector | None = None


def get_text_run_type(v: object | str | dict[str, Any]) -> str:
    if isinstance(v, str):
        return "str"
    if isinstance(v, dict) and "propertyIdentifier" in v:
        return "PropertyIdentifier"
    return v.__class__.__name__


TextRunValue = Annotated[
    Annotated[str, Tag("str")] | Annotated[PropertyIdentifier, Tag("PropertyIdentifier")],
    Discriminator(get_text_run_type),
]


class TextRun(LayoutNode):
    textStyle: TextStyle | None = None
    value: TextRunValue | None = None
    cases: list[Case] | None = None
    defaultCase: DefaultCase | None = None
    url: str | None = None
    expression: TextRunExpression | None = None


class Paragraph(LayoutNode):
    horizontalTextAlignment: str | None = None  # TODO: convert to enum
    textRuns: list[TextRun]



import datetime
from enum import IntEnum, StrEnum
from typing import TYPE_CHECKING, Any, Literal

import pydantic
from pydantic_extra_types.semantic_version import SemanticVersion

from pbi_core.lineage import LineageNode
from pbi_core.pydantic import BaseValidation
from pbi_core.ssas.model_tables.base import SsasEditableRecord
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.culture import Culture


class ContentType(IntEnum):
    Xml = 0
    Json = 1


class EntityDefinitionBinding(BaseValidation):
    ConceptualEntity: str
    ConceptualProperty: str | None = None
    VariationSource: str | None = None
    VariationSet: str | None = None
    Hierarchy: str | None = None
    HierarchyLevel: str | None = None


class EntityDefinition(BaseValidation):
    Binding: EntityDefinitionBinding


class TermSourceType(StrEnum):
    External = "External"


class TermSource(BaseValidation):
    Type: TermSourceType | None = None
    Agent: str


class TermDefinitionState(StrEnum):
    Suggested = "Suggested"
    Generated = "Generated"
    Deleted = "Deleted"


class TermDefinitionType(StrEnum):
    Noun = "Noun"


class TermDefinition(BaseValidation):
    State: TermDefinitionState | None = None
    Source: TermSource | None = None
    Weight: float | None = None
    Type: TermDefinitionType | None = None
    LastModified: datetime.datetime | None = None


class VisibilityValue(StrEnum):
    Hidden = "Hidden"


class VisibilityState(StrEnum):
    Authored = "Authored"


class VisibilityType(BaseValidation):
    Value: VisibilityValue
    State: VisibilityState | None = None


class NameTypeType(StrEnum):
    Identifier = "Identifier"
    Name = "Name"


class LinguisticMetadataEntity(BaseValidation):
    Weight: float | None = None
    State: TermDefinitionState
    Terms: list[dict[str, TermDefinition]] | None = None
    Definition: EntityDefinition | None = None
    Binding: EntityDefinitionBinding | None = None
    SemanticType: str | None = None
    Visibility: VisibilityType | None = None
    Hidden: bool = False
    NameType: NameTypeType | None = None
    Units: list[str] = []


class RelationshipBinding(BaseValidation):
    ConceptualEntity: str


class PhrasingAttributeRole(BaseValidation):
    Role: str


class RelationshipPhrasingState(StrEnum):
    Generated = "Generated"


# TODO: Subtype
class PhrasingAttribute(BaseValidation):
    model_config = pydantic.ConfigDict(
        extra="forbid",
    )
    Adjective: PhrasingAttributeRole | None = None
    Measurement: PhrasingAttributeRole | None = None
    Object: PhrasingAttributeRole | None = None
    Subject: PhrasingAttributeRole | None = None
    Name: PhrasingAttributeRole | None = None

    PrepositionalPhrases: list[dict[str, Any]] = []
    Adjectives: list[dict[str, TermDefinition]] = []
    Antonyms: list[dict[str, TermDefinition]] = []
    Prepositions: list[dict[str, TermDefinition]] = []
    Verbs: list[dict[str, TermDefinition]] = []
    Nouns: list[dict[str, TermDefinition]] = []


class RelationshipPhrasing(BaseValidation):
    Name: PhrasingAttribute | None = None
    Attribute: PhrasingAttribute | None = None
    Verb: PhrasingAttribute | None = None
    Adjective: PhrasingAttribute | None = None
    Preposition: PhrasingAttribute | None = None
    DynamicAdjective: PhrasingAttribute | None = None
    State: RelationshipPhrasingState | None = None
    Weight: float | None = None


class RelationshipRoleEntity(BaseValidation):
    Entity: str


class RelationshipRole(BaseValidation):
    Target: RelationshipRoleEntity
    Nouns: Any | None = None


class SemanticSlot(BaseValidation):
    Where: PhrasingAttributeRole | None = None
    When: PhrasingAttributeRole | None = None


class ConditionOperator(StrEnum):
    Equals = "Equals"
    GreaterThan = "GreaterThan"


class Condition(BaseValidation):
    Target: PhrasingAttributeRole
    Operator: ConditionOperator
    Value: dict[str, list[int | str]]


class LinguisticMetadataRelationship(BaseValidation):
    Binding: RelationshipBinding
    Phrasings: list[RelationshipPhrasing] = []
    Roles: dict[str, RelationshipRole | int]
    State: str | None = None
    SemanticSlots: SemanticSlot | None = None
    Conditions: list[Condition] | None = None


class LinguisticMetadataContent(BaseValidation):
    Version: SemanticVersion
    Language: str
    DynamicImprovement: str | None = None
    Relationships: dict[str, LinguisticMetadataRelationship] | None = None
    Entities: dict[str, LinguisticMetadataEntity] | None = None
    Examples: list[dict[str, dict[str, str]]] | None = None


class LinguisticMetadata(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/f8924a45-70da-496a-947a-84b8d5beaae6)
    """

    content: pydantic.Json[LinguisticMetadataContent]
    content_type: ContentType
    culture_id: int

    modified_time: datetime.datetime

    _commands: BaseCommands = pydantic.PrivateAttr(default_factory=lambda: SsasCommands.linguistic_metadata)

    def culture(self) -> "Culture":
        return self.tabular_model.cultures.find({"id": self.culture_id})

    def pbi_core_name(self) -> str:
        """Returns the name displayed in the PBIX report."""
        return self.culture().pbi_core_name()

    @classmethod
    def _db_command_obj_name(cls) -> str:
        return "LinguisticMetadata"

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(self, lineage_type)
        return LineageNode(self, lineage_type, [self.culture().get_lineage(lineage_type)])



from ._base_node import LayoutNode
from .condition import ConditionType


class DataViewWildcard(LayoutNode):
    matchingOption: int
    roles: list[str] | None = None


class SelectorData(LayoutNode):
    roles: list[str] | None = None
    dataViewWildcard: DataViewWildcard | None = None
    scopeId: ConditionType | None = None


# TODO: possibly replace with a union?
class Selector(LayoutNode):
    id: str | None = None
    # Weird values, pretty confident this is not an enum
    metadata: str | None = None
    data: list[SelectorData] | None = None



from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual
from .column_property import ColumnProperty
from .properties.base import Expression


class CategoryAxisPropertiesHelper(LayoutNode):
    axisType: Expression | None = None
    concatenateLabels: Expression | None = None
    end: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    gridlineShow: Expression | None = None
    labelColor: Expression | None = None
    maxMarginFactor: Expression | None = None
    show: Expression | None = None
    showAxisTitle: Expression | None = None
    start: Expression | None = None
    titleColor: Expression | None = None
    titleFontFamily: Expression | None = None
    titleFontSize: Expression | None = None


class CategoryAxisProperties(LayoutNode):
    properties: CategoryAxisPropertiesHelper
    selector: Selector | None = None


class LineStylesPropertiesHelper(LayoutNode):
    lineStyle: Expression | None = None
    markerColor: Expression | None = None
    markerShape: Expression | None = None
    markerSize: Expression | None = None
    showMarker: Expression | None = None
    showSeries: Expression | None = None
    stepped: Expression | None = None
    strokeLineJoin: Expression | None = None
    strokeWidth: Expression | None = None


class LineStylesProperties(LayoutNode):
    properties: LineStylesPropertiesHelper
    selector: Selector | None = None


class LabelsPropertiesHelper(LayoutNode):
    color: Expression | None = None
    labelPosition: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    labelDensity: Expression | None = None
    show: Expression | None = None
    showAll: Expression | None = None
    showSeries: Expression | None = None


class LabelsProperties(LayoutNode):
    properties: LabelsPropertiesHelper
    selector: Selector | None = None


class DataPointPropertiesHelper(LayoutNode):
    fill: Expression | None = None
    showAllDataPoints: Expression | None = None


class DataPointProperties(LayoutNode):
    properties: DataPointPropertiesHelper
    selector: Selector | None = None


class LegendPropertiesHelper(LayoutNode):
    defaultToCircle: Expression | None = None
    fontSize: Expression | None = None
    labelColor: Expression | None = None
    legendMarkerRendering: Expression | None = None
    position: Expression | None = None
    show: Expression | None = None
    showTitle: Expression | None = None
    titleText: Expression | None = None


class LegendProperties(LayoutNode):
    properties: LegendPropertiesHelper


class ValueAxisPropertiesHelper(LayoutNode):
    axisScale: Expression | None = None
    end: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    gridlineColor: Expression | None = None
    gridlineShow: Expression | None = None
    gridlineStyle: Expression | None = None
    gridlineThickness: Expression | None = None
    labelColor: Expression | None = None
    labelDensity: Expression | None = None
    labelDisplayUnits: Expression | None = None
    position: Expression | None = None
    show: Expression | None = None
    showAxisTitle: Expression | None = None
    start: Expression | None = None
    titleFontFamily: Expression | None = None
    titleText: Expression | None = None


class ValueAxisProperties(LayoutNode):
    properties: ValueAxisPropertiesHelper


class AnomalyDetectionPropertiesHelper(LayoutNode):
    confidenceBandColor: Expression | None = None
    displayName: Expression | None = None
    explainBy: Expression | None = None
    markerColor: Expression | None = None
    markerShape: Expression | None = None
    show: Expression | None = None
    transform: Expression | None = None
    transparency: Expression | None = None


class AnomalyDetectionProperties(LayoutNode):
    properties: AnomalyDetectionPropertiesHelper
    selector: Selector | None = None


class ZoomPropertiesHelper(LayoutNode):
    show: Expression | None = None
    categoryMax: Expression | None = None
    categoryMin: Expression | None = None
    valueMax: Expression | None = None
    valueMin: Expression | None = None


class ZoomProperties(LayoutNode):
    properties: ZoomPropertiesHelper


class ForecastPropertiesHelper(LayoutNode):
    show: Expression | None = None
    displayName: Expression | None = None
    lineColor: Expression | None = None
    transform: Expression | None = None


class ForecastProperties(LayoutNode):
    properties: ForecastPropertiesHelper
    selector: Selector | None = None


class PlotAreaPropertiesHelper(LayoutNode):
    transparency: Expression | None = None


class PlotAreaProperties(LayoutNode):
    properties: PlotAreaPropertiesHelper


class TrendPropertiesHelper(LayoutNode):
    displayName: Expression | None = None
    lineColor: Expression | None = None
    show: Expression | None = None


class TrendProperties(LayoutNode):
    properties: TrendPropertiesHelper


class GeneralPropertiesHelper(LayoutNode):
    responsive: Expression | None = None


class GeneralProperties(LayoutNode):
    properties: GeneralPropertiesHelper


class Y2AxisPropertiesHelper(LayoutNode):
    show: Expression | None = None


class Y2AxisProperties(LayoutNode):
    properties: Y2AxisPropertiesHelper


class Y1AxisReferenceLinePropertiesHelper(LayoutNode):
    displayName: Expression | None = None
    position: Expression | None = None
    show: Expression | None = None
    style: Expression | None = None


class Y1AxisReferenceLineProperties(LayoutNode):
    properties: Y1AxisReferenceLinePropertiesHelper
    selector: Selector | None = None


class LineChartProperties(LayoutNode):
    anomalyDetection: list[AnomalyDetectionProperties] | None = None
    categoryAxis: list[CategoryAxisProperties] | None = None
    dataPoint: list[DataPointProperties] | None = None
    forecast: list[ForecastProperties] | None = None
    general: list[GeneralProperties] | None = None
    labels: list[LabelsProperties] | None = None
    legend: list[LegendProperties] | None = None
    lineStyles: list[LineStylesProperties] | None = None
    plotArea: list[PlotAreaProperties] | None = None
    trend: list[TrendProperties] | None = None
    valueAxis: list[ValueAxisProperties] | None = None
    zoom: list[ZoomProperties] | None = None
    y1AxisReferenceLine: list[Y1AxisReferenceLineProperties] | None = None
    y2Axis: list[Y2AxisProperties] | None = None


class LineChart(BaseVisual):
    visualType: str = "lineChart"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: LineChartProperties | None = None



from typing import Any

from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector
from pbi_core.static_files.layout.sources.paragraphs import Paragraph

from .base import BaseVisual
from .properties.base import Expression


class GeneralProperties(LayoutNode):
    paragraphs: list[Paragraph] | None = None
    responsive: Expression | None = None


class General(LayoutNode):
    properties: GeneralProperties


class ValuePropertiesHelper(LayoutNode):
    context: Any | None = None  # TODO: should be Source, but causes circular import issues with Subquery
    expr: Any | None = None  # TODO: should be Source, but causes circular import issues with Subquery
    value: Any | None = None  # TODO: should be Source, but causes circular import issues with Subquery
    propertyDefinitionKind: str | None = None


class ValueProperties(LayoutNode):
    expr: ValuePropertiesHelper
    formatString: Expression | None = None


class Value(LayoutNode):
    properties: ValueProperties | None = None
    selector: Selector | None = None


class TextBoxProperties(LayoutNode):
    general: list[General]
    values: list[Value] | None = None


class TextBox(BaseVisual):
    visualType: str = "textbox"
    model_config = ConfigDict(extra="forbid")

    drillFilterOtherVisuals: bool = True
    objects: TextBoxProperties | None = None



from typing import Literal

from pbi_parsers import dax

from .enums import Alignment, ColumnType
from .relationships import RelationshipMixin


class HelpersMixin(RelationshipMixin):
    alignment: Alignment
    expression: str | int | None
    system_flags: int
    explicit_name: str | None
    id: int
    inferred_name: str | None
    type: ColumnType

    def expression_ast(self) -> dax.Expression | None:
        if not isinstance(self.expression, str):
            return None
        return dax.to_ast(self.expression)

    def is_system_table(self) -> bool:
        return bool(self.system_flags >> 1 % 2)

    def is_from_calculated_table(self) -> bool:
        return bool(self.system_flags % 2)

    def is_calculated_column(self) -> bool:
        return self.type == ColumnType.CALCULATED

    def is_normal(self) -> bool:
        """Returns True if the column is not a row number column or attached to a system/private table."""
        if self.type == ColumnType.ROW_NUMBER:
            return False
        if self.table().is_private:
            return False
        return not self.is_system_table()

    def pbi_core_name(self) -> str:
        """Returns the name displayed in the PBIX report."""
        if self.explicit_name is not None:
            return self.explicit_name
        if self.inferred_name is not None:
            return self.inferred_name
        return str(self.id)

    def _column_type(self) -> Literal["COLUMN", "CALC_COLUMN"]:
        if self.type == ColumnType.DATA:
            return "COLUMN"
        return "CALC_COLUMN"

    def modification_hash(self) -> int:
        return hash((
            self.alignment,
            self.explicit_name,
            self.sort_by_column_id,
            self.table_id,
        ))

    def full_name(self) -> str:
        """Returns the fully qualified name for DAX queries.

        Examples:
            'TableName'[ColumnName]

        """
        table_name = self.table().name
        return f"'{table_name}'[{self.explicit_name}]"

    def data(self, head: int = 100) -> list[int | float | str]:
        table_name = self.table().name
        ret = self.tabular_model.server.query_dax(
            f"EVALUATE TOPN({head}, SELECTCOLUMNS(ALL('{table_name}'), {self.full_name()}))",
            db_name=self.tabular_model.db_name,
        )
        return [next(iter(row.values())) for row in ret]



from enum import IntEnum
from typing import Annotated, Any

from pydantic import Discriminator, Tag

from pbi_core.static_files.layout._base_node import LayoutNode


class EntityType(IntEnum):
    NA = 1
    NA2 = 0
    NA3 = 2


class Entity(LayoutNode):
    Entity: str
    Name: str | None = None
    Type: EntityType | None = EntityType.NA2

    def table(self) -> str:
        return self.Entity

    def table_mapping(self) -> dict[str, str]:
        if self.Name is None:
            return {}
        return {self.Name: self.Entity}

    @staticmethod
    def create(entity: str) -> "Entity":
        return Entity.model_validate({"Entity": entity})

    def __repr__(self) -> str:
        return f"Entity({self.Name}: {self.Entity})"


class Source(LayoutNode):
    Source: str

    def table(self) -> str:
        return self.Source


def get_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "Source" in v:
            return "Source"
        if "Entity" in v:
            return "Entity"
        raise TypeError(v)
    return v.__class__.__name__


SourceRefSource = Annotated[
    Annotated[Entity, Tag("Entity")] | Annotated[Source, Tag("Source")],
    Discriminator(get_type),
]


class TransformTableRef(LayoutNode):
    TransformTableRef: SourceRefSource

    def table(self) -> str:
        return self.TransformTableRef.table()


class SourceRef(LayoutNode):
    SourceRef: SourceRefSource

    def table(self) -> str:
        return self.SourceRef.table()


def get_source_expression_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "SourceRef" in v:
            return "SourceRef"
        if "TransformTableRef" in v:
            return "TransformTableRef"
        raise TypeError(v)
    return v.__class__.__name__


SourceExpressionUnion = Annotated[
    Annotated[TransformTableRef, Tag("TransformTableRef")] | Annotated[SourceRef, Tag("SourceRef")],
    Discriminator(get_source_expression_type),
]


class SourceExpression(LayoutNode):
    Expression: SourceExpressionUnion
    Property: str

    def table(self) -> str:
        return self.Expression.table()

    def column(self) -> str:
        return self.Property

    @staticmethod
    def create(table: str, column: str) -> "SourceExpression":
        entity = Entity.create(entity=table)
        ret: SourceExpression = SourceExpression.model_validate({
            "Expression": {
                "SourceRef": entity.model_dump_json(),
                "Property": column,
            },
        })
        return ret



from .connections import Connections
from .diagram_layout import DiagramLayout, parse_diagram_layout
from .metadata import Metadata
from .settings import Settings

__all__ = ["Connections", "DiagramLayout", "Metadata", "Settings", "parse_diagram_layout"]



import datetime
from typing import TYPE_CHECKING, Literal

from pbi_core.lineage import LineageNode
from pbi_core.ssas.model_tables.base import SsasReadonlyRecord
from pbi_core.ssas.model_tables.enums import DataState

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.column import Column
    from pbi_core.ssas.model_tables.level import Level


class AttributeHierarchy(SsasReadonlyRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/93d1844f-a6c7-4dda-879b-2e26ed5cd297)
    """

    attribute_hierarchy_storage_id: int
    column_id: int
    state: DataState

    modified_time: datetime.datetime
    refreshed_time: datetime.datetime

    def pbi_core_name(self) -> str:
        """Returns the name displayed in the PBIX report."""
        return self.column().pbi_core_name()

    def column(self) -> "Column":
        return self.tabular_model.columns.find({"id": self.column_id})

    def levels(self) -> set["Level"]:
        return self.tabular_model.levels.find_all({"hierarchy_id": self.id})

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(self, lineage_type, [level.get_lineage(lineage_type) for level in self.levels()])
        return LineageNode(self, lineage_type, [self.column().get_lineage(lineage_type)])



from bs4 import BeautifulSoup
from structlog import BoundLogger

from pbi_core.logging import get_logger

from .dependencies import DependencyMixin

logger: BoundLogger = get_logger()


class CommandMixin(DependencyMixin):
    def delete(
        self,
        *,
        remote_from_partition_query: bool = False,
    ) -> list[BeautifulSoup] | None:
        """Removes column from the SSAS model.

        Args:
            remote_from_partition_query (bool): If this is True, updates the PowerQuery to drop the column so it doesn't repopulate on data refresh. If the column is calculated (meaning it is not part of the PowerQuery), an info record is logged and nothing else happens.

        """  # noqa: E501
        rets: list[BeautifulSoup] = []
        if remote_from_partition_query:
            if self._column_type() == "CALC_COLUMN":
                logger.info("Column is calculated, there is nothing to remove from the PowerQuery", column=self)
                return None
            name = self.explicit_name
            if name is None:
                logger.warning("Column has no name to include in the PowerQuery", column=self)
                return None
            rets.extend(partition.remove_columns([name]) for partition in self.table().partitions())
        rets.append(super().delete())  # pyright: ignore[reportAttributeAccessIssue]
        return rets



import datetime
from typing import TYPE_CHECKING

from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasRenameRecord
from pbi_core.ssas.server._commands import RenameCommands
from pbi_core.ssas.server.utils import SsasCommands

from .enums import DataSourceType, ImpersonationMode, Isolation

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.model import Model


class DataSource(SsasRenameRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/ee12dcb7-096e-4e4e-99a4-47caeb9390f5)
    """

    account: str | None = None
    connection_string: str
    context_expression: str | None = None
    credential: str | None = None
    description: str | None = None
    impersonation_mode: ImpersonationMode
    isolation: Isolation
    max_connections: int
    model_id: int
    name: str
    options: str | None = None
    password: str | None = None
    provider: str | None = None
    timeout: int
    type: DataSourceType

    modified_time: datetime.datetime

    _commands: RenameCommands = PrivateAttr(default_factory=lambda: SsasCommands.data_source)

    def model(self) -> "Model":
        return self.tabular_model.model



import typing

if typing.TYPE_CHECKING:
    from pbi_core.ssas.model_tables.base import SsasTable

from ._group import Group, RowNotFoundError
from .alternate_of import AlternateOf
from .annotation import Annotation
from .attribute_hierarchy import AttributeHierarchy
from .calc_dependency import CalcDependency
from .calculation_group import CalculationGroup
from .calculation_item import CalculationItem
from .column import Column
from .column_permission import ColumnPermission
from .culture import Culture
from .data_source import DataSource
from .detail_row_definition import DetailRowDefinition
from .expression import Expression
from .extended_property import ExtendedProperty
from .format_string_definition import FormatStringDefinition
from .group_by_column import GroupByColumn
from .hierarchy import Hierarchy
from .kpi import KPI
from .level import Level
from .linguistic_metadata import LinguisticMetadata
from .measure import Measure
from .model import Model
from .object_translation import ObjectTranslation
from .partition import Partition
from .perspective import Perspective
from .perspective_column import PerspectiveColumn
from .perspective_hierarchy import PerspectiveHierarchy
from .perspective_measure import PerspectiveMeasure
from .perspective_set import PerspectiveSet
from .perspective_table import PerspectiveTable
from .query_group import QueryGroup
from .refresh_policy import RefreshPolicy
from .related_column_detail import RelatedColumnDetail
from .relationship import Relationship
from .role import Role
from .role_membership import RoleMembership
from .set import Set
from .table import Table
from .table_permission import TablePermission
from .variation import Variation

__all__ = [
    "KPI",
    "AlternateOf",
    "Annotation",
    "AttributeHierarchy",
    "CalcDependency",
    "CalculationGroup",
    "CalculationItem",
    "Column",
    "ColumnPermission",
    "Culture",
    "DataSource",
    "DetailRowDefinition",
    "Expression",
    "ExtendedProperty",
    "FormatStringDefinition",
    "Group",
    "GroupByColumn",
    "Hierarchy",
    "Level",
    "LinguisticMetadata",
    "Measure",
    "Model",
    "ObjectTranslation",
    "Partition",
    "Perspective",
    "PerspectiveColumn",
    "PerspectiveHierarchy",
    "PerspectiveMeasure",
    "PerspectiveSet",
    "PerspectiveTable",
    "QueryGroup",
    "RefreshPolicy",
    "RelatedColumnDetail",
    "Relationship",
    "Role",
    "RoleMembership",
    "RowNotFoundError",
    "Set",
    "Table",
    "TablePermission",
    "Variation",
]

FIELD_TYPES: dict[str, type["SsasTable"]] = {
    "alternate_ofs": AlternateOf,
    "annotations": Annotation,
    "attribute_hierarchies": AttributeHierarchy,
    "calc_dependencies": CalcDependency,
    "calculation_groups": CalculationGroup,
    "calculation_items": CalculationItem,
    "column_permissions": ColumnPermission,
    "columns": Column,
    "cultures": Culture,
    "data_sources": DataSource,
    "detail_row_definitions": DetailRowDefinition,
    "expressions": Expression,
    "extended_properties": ExtendedProperty,
    "format_string_definitions": FormatStringDefinition,
    "group_by_columns": GroupByColumn,
    "hierarchies": Hierarchy,
    "kpis": KPI,
    "levels": Level,
    "linguistic_metadata": LinguisticMetadata,
    "measures": Measure,
    "model": Model,
    "object_translations": ObjectTranslation,
    "partitions": Partition,
    "perspective_columns": PerspectiveColumn,
    "perspective_hierarchies": PerspectiveHierarchy,
    "perspective_measures": PerspectiveMeasure,
    "perspective_sets": PerspectiveSet,
    "perspective_tables": PerspectiveTable,
    "perspectives": Perspective,
    "query_groups": QueryGroup,
    "relationships": Relationship,
    "refresh_policies": RefreshPolicy,
    "related_column_details": RelatedColumnDetail,
    "role_memberships": RoleMembership,
    "roles": Role,
    "sets": Set,
    "table_permissions": TablePermission,
    "tables": Table,
    "variations": Variation,
}



import click

from pbi_core.misc import remove_workspaces
from pbi_core.ssas.setup import interactive_setup


@click.group()
def pbi_core_commands() -> None:
    pass


@pbi_core_commands.command()
@click.option(
    "--plan",
    is_flag=True,
    show_default=True,
    default=False,
    help="If set, displays the workspaces to be deleted",
)
def clean(*, plan: bool) -> None:
    """Removes all workspaces left over by the library."""
    if plan:
        remove_workspaces.list_workspaces()
    else:
        remove_workspaces.clear_workspaces()


@pbi_core_commands.command()
def setup() -> None:
    """Interactive setup for package dependencies.

    Note: msmdsrv.ini auto-find requires PowerBI Desktop to be running.
    """
    interactive_setup()


if __name__ == "__main__":
    pbi_core_commands()



# C:\Program Files\Microsoft SQL Server\MSAS16.MSSQLSERVER\OLAP\bin\Resources\1033\tracedefinition.xml
# ruff: noqa: E501
from enum import IntEnum

from .events.command_events import CommandBeginColumns, CommandEndColumns
from .events.discover_events import DiscoverBeginColumns, DiscoverEndColumns
from .events.discover_server_states import (
    ServerStateDiscoverBeginColumns,
    ServerStateDiscoverDataColumns,
    ServerStateDiscoverEndColumns,
)
from .events.errors_and_warnings import ErrorColumns
from .events.file_load_and_save import (
    FileLoadBeginColumns,
    FileLoadEndColumns,
    FileSaveBeginColumns,
    FileSaveEndColumns,
    PageInBeginColumns,
    PageInEndColumns,
    PageOutBeginColumns,
    PageOutEndColumns,
)
from .events.locks import (
    DeadlockColumns,
    LockAcquiredColumns,
    LockReleasedColumns,
    LockTimeoutColumns,
    LockWaitingColumns,
)
from .events.m_data_provider_events import ExecuteSourceQueryColumns
from .events.notifications import NotificationColumns, UserDefinedColumns
from .events.progress_reports import (
    ProgressReportBeginColumns,
    ProgressReportCurrentColumns,
    ProgressReportEndColumns,
    ProgressReportErrorColumns,
)
from .events.query_events import QueryBeginColumns, QueryEndColumns
from .events.query_processing import (
    AggregateTableRewriteInfoColumns,
    AggregateTableRewriteQueryColumns,
    CalculationEvaluationColumns,
    CalculationEvaluationDetailedInformationColumns,
    DaxExtensionExecutionBeginColumns,
    DaxExtensionExecutionEndColumns,
    DaxExtensionTraceErrorColumns,
    DaxExtensionTraceInfoColumns,
    DaxExtensionTraceVerboseColumns,
    DaxQueryPlanColumns,
    DirectQueryBeginColumns,
    DirectQueryEndColumns,
    ExecutionMetricsColumns,
    QuerySubcubeColumns,
    VertipaqSeQueryBeginColumns,
    VertipaqSeQueryCacheMatchColumns,
    VertipaqSeQueryCacheMissColumns,
    VertipaqSeQueryEndColumns,
)
from .events.resource_governance import (
    WlgroupCpuThrottlingColumns,
    WlgroupExceedsMemoryLimitColumns,
    WlgroupExceedsProcessingLimitColumns,
)
from .events.security_audit import (
    AuditAdminOperationsEventColumns,
    AuditLoginColumns,
    AuditLogoutColumns,
    AuditObjectPermissionEventColumns,
    AuditServerStartsAndStopsColumns,
)
from .events.sessions import ExistingConnectionColumns, ExistingSessionColumns, SessionInitializeColumns


class TraceEvents(IntEnum):
    AUDIT_LOGIN = 1  # Collects all new connection events since the trace was started, such as when a client requests a connection to a server running an instance of SQL Server.
    AUDIT_LOGOUT = 2  # Collects all new disconnect events since the trace was started, such as when a client issues a disconnect command.
    AUDIT_SERVER_STARTS_AND_STOPS = 4  # Records service shut down, start, and pause activities.

    PROGRESS_REPORT_BEGIN = 5  # Progress report begins.
    PROGRESS_REPORT_END = 6  # Progress report end.
    PROGRESS_REPORT_CURRENT = 7  # Progress report current.
    PROGRESS_REPORT_ERROR = 8  # Progress report error.
    QUERY_BEGIN = 9  # Query begins.
    QUERY_END = 10  # Query end.
    QUERY_SUBCUBE = 11  # Query subcube, for Usage Based Optimization.
    QUERY_SUBCUBE_VERBOSE = 12  # Query subcube with detailed information. This event may have a negative impact on performance when turned on.

    COMMAND_BEGIN = 15  # Command begin.
    COMMAND_END = 16  # Command end.
    ERROR = 17  # Server error.
    AUDIT_OBJECT_PERMISSION_EVENT = 18  # Records object permission changes.
    AUDIT_ADMIN_OPERATIONS_EVENT = 19  # Records server backup/restore/synchronize/attach/detach/imageload/imagesave.

    SERVER_STATE_DISCOVER_BEGIN = 33  # Start of Server State Discover.
    SERVER_STATE_DISCOVER_DATA = 34  # Contents of the Server State Discover Response.
    SERVER_STATE_DISCOVER_END = 35  # End of Server State Discover.
    DISCOVER_BEGIN = 36  # Start of Discover Request.
    DISCOVER_END = 38  # End of Discover Request.
    NOTIFICATION = 39  # Notification event.
    USER_DEFINED = 40  # User defined Event.
    EXISTING_CONNECTION = 41  # Existing user connection.
    EXISTING_SESSION = 42  # Existing session.
    SESSION_INITIALIZE = 43  # Session Initialize.

    DEADLOCK = 50  # Metadata locks deadlock.
    LOCK_TIMEOUT = 51  # Metadata lock timeout.
    LOCK_ACQUIRED = 52  # Lock Acquired
    LOCK_RELEASED = 53  # Lock Released
    LOCK_WAITING = 54  # Lock Waiting
    GET_DATA_FROM_AGGREGATION = 60  # Answer query by getting data from aggregation. This event may have a negative impact on performance when turned on.
    GET_DATA_FROM_CACHE = 61  # Answer query by getting data from one of the caches. This event may have a negative impact on performance when turned on.

    QUERY_CUBE_BEGIN = 70  # Query cube begin.
    QUERY_CUBE_END = 71  # Query cube end.
    CALCULATE_NON_EMPTY_BEGIN = 72  # Calculate non empty begin.
    CALCULATE_NON_EMPTY_CURRENT = 73  # Calculate non empty current.
    CALCULATE_NON_EMPTY_END = 74  # Calculate non empty end.
    SERIALIZE_RESULTS_BEGIN = 75  # Serialize results begin.
    SERIALIZE_RESULTS_CURRENT = 76  # Serialize results current.
    SERIALIZE_RESULTS_END = 77  # Serialize results end.
    EXECUTE_MDX_SCRIPT_BEGIN = 78  # Execute MDX script begin.
    EXECUTE_MDX_SCRIPT_CURRENT = 79  # Execute MDX script current. Deprecated.
    EXECUTE_MDX_SCRIPT_END = 80  # Execute MDX script end.
    QUERY_DIMENSION = 81  # Query dimension.
    VERTIPAQ_SE_QUERY_BEGIN = 82  # VertiPaq SE query
    VERTIPAQ_SE_QUERY_END = 83  # VertiPaq SE query
    RESOURCE_USAGE = 84  # Reports reads, writes, cpu usage after end of commands and queries.
    VERTIPAQ_SE_QUERY_CACHE_MATCH = 85  # VertiPaq SE query cache use
    VERTIPAQ_SE_QUERY_CACHE_MISS = 86  # VertiPaq SE Query Cache Miss

    FILE_LOAD_BEGIN = 90  # File Load Begin.
    FILE_LOAD_END = 91  # File Load End.
    FILE_SAVE_BEGIN = 92  # File Save Begin.
    FILE_SAVE_END = 93  # File Save End
    PAGEOUT_BEGIN = 94  # PageOut Begin.
    PAGEOUT_END = 95  # PageOut End
    PAGEIN_BEGIN = 96  # PageIn Begin.
    PAGEIN_END = 97  # PageIn End
    DIRECT_QUERY_BEGIN = 98  # Direct Query Begin.
    DIRECT_QUERY_END = 99  # Direct Query End.

    CALCULATION_EVALUATION = 110  # Information about the evaluation of calculations. This event will have a negative impact on performance when turned on.
    CALCULATION_EVALUATION_DETAILED_INFORMATION = 111  # Detailed information about the evaluation of calculations. This event will have a negative impact on performance when turned on.
    DAX_QUERY_PLAN = 112  # DAX logical/physical plan tree for VertiPaq and DirectQuery modes.
    WLGROUP_CPU_THROTTLING = 113  # Workload Group is throttled on CPU usage
    WLGROUP_EXCEEDS_MEMORY_LIMIT = 114  # Workload group exceeds the memory limit
    WLGROUP_EXCEEDS_PROCESSING_LIMIT = 115  # Workload group exceeds the processing limit
    DAX_EXTENSION_EXECUTION_BEGIN = 120  # DAX extension function execution begin event.
    DAX_EXTENSION_EXECUTION_END = 121  # DAX extension function execution end event.
    DAX_EXTENSION_TRACE_ERROR = 122  # DAX extension function error trace event directly traced by extension authors.
    DAX_EXTENSION_TRACE_INFO = (
        123  # DAX extension function informational/telemetry trace event directly traced by extension authors.
    )
    DAX_EXTENSION_TRACE_VERBOSE = (
        124  # DAX extension function verbose trace event directly traced by extension authors.
    )
    EXECUTE_SOURCE_QUERY = 130  # Collection of all queries that are executed against the data source
    AGGREGATE_TABLE_REWRITE_QUERY = 131  # A query was rewritten according to available aggregate tables
    AGGREGATE_TABLE_REWRITE_INFO = 132  # Aggregate Table Rewrite Info

    JOB_GRAPH = 134  # Collection of Job Graph related events.
    EXECUTION_METRICS = 136

    def get_columns(self) -> type[IntEnum]:
        return event_column_mapping[self]


event_column_mapping: dict[TraceEvents, type[IntEnum]] = {
    TraceEvents.AGGREGATE_TABLE_REWRITE_INFO: AggregateTableRewriteInfoColumns,
    TraceEvents.AGGREGATE_TABLE_REWRITE_QUERY: AggregateTableRewriteQueryColumns,
    TraceEvents.AUDIT_ADMIN_OPERATIONS_EVENT: AuditAdminOperationsEventColumns,
    TraceEvents.AUDIT_LOGIN: AuditLoginColumns,
    TraceEvents.AUDIT_LOGOUT: AuditLogoutColumns,
    TraceEvents.AUDIT_OBJECT_PERMISSION_EVENT: AuditObjectPermissionEventColumns,
    TraceEvents.AUDIT_SERVER_STARTS_AND_STOPS: AuditServerStartsAndStopsColumns,
    TraceEvents.CALCULATION_EVALUATION: CalculationEvaluationColumns,
    TraceEvents.CALCULATION_EVALUATION_DETAILED_INFORMATION: CalculationEvaluationDetailedInformationColumns,
    TraceEvents.COMMAND_BEGIN: CommandBeginColumns,
    TraceEvents.COMMAND_END: CommandEndColumns,
    TraceEvents.DAX_EXTENSION_EXECUTION_BEGIN: DaxExtensionExecutionBeginColumns,
    TraceEvents.DAX_EXTENSION_EXECUTION_END: DaxExtensionExecutionEndColumns,
    TraceEvents.DAX_EXTENSION_TRACE_ERROR: DaxExtensionTraceErrorColumns,
    TraceEvents.DAX_EXTENSION_TRACE_INFO: DaxExtensionTraceInfoColumns,
    TraceEvents.DAX_EXTENSION_TRACE_VERBOSE: DaxExtensionTraceVerboseColumns,
    TraceEvents.DIRECT_QUERY_BEGIN: DirectQueryBeginColumns,
    TraceEvents.DIRECT_QUERY_END: DirectQueryEndColumns,
    TraceEvents.DAX_QUERY_PLAN: DaxQueryPlanColumns,
    TraceEvents.DEADLOCK: DeadlockColumns,
    TraceEvents.DISCOVER_BEGIN: DiscoverBeginColumns,
    TraceEvents.DISCOVER_END: DiscoverEndColumns,
    TraceEvents.ERROR: ErrorColumns,
    TraceEvents.EXISTING_CONNECTION: ExistingConnectionColumns,
    TraceEvents.EXISTING_SESSION: ExistingSessionColumns,
    TraceEvents.EXECUTE_SOURCE_QUERY: ExecuteSourceQueryColumns,
    TraceEvents.EXECUTION_METRICS: ExecutionMetricsColumns,
    TraceEvents.FILE_SAVE_BEGIN: FileSaveBeginColumns,
    TraceEvents.FILE_SAVE_END: FileSaveEndColumns,
    TraceEvents.FILE_LOAD_BEGIN: FileLoadBeginColumns,
    TraceEvents.FILE_LOAD_END: FileLoadEndColumns,
    TraceEvents.LOCK_ACQUIRED: LockAcquiredColumns,
    TraceEvents.LOCK_RELEASED: LockReleasedColumns,
    TraceEvents.LOCK_TIMEOUT: LockTimeoutColumns,
    TraceEvents.LOCK_WAITING: LockWaitingColumns,
    TraceEvents.NOTIFICATION: NotificationColumns,
    TraceEvents.PAGEIN_BEGIN: PageInBeginColumns,
    TraceEvents.PAGEIN_END: PageInEndColumns,
    TraceEvents.PAGEOUT_BEGIN: PageOutBeginColumns,
    TraceEvents.PAGEOUT_END: PageOutEndColumns,
    TraceEvents.PROGRESS_REPORT_BEGIN: ProgressReportBeginColumns,
    TraceEvents.PROGRESS_REPORT_CURRENT: ProgressReportCurrentColumns,
    TraceEvents.PROGRESS_REPORT_END: ProgressReportEndColumns,
    TraceEvents.PROGRESS_REPORT_ERROR: ProgressReportErrorColumns,
    TraceEvents.QUERY_BEGIN: QueryBeginColumns,
    TraceEvents.QUERY_END: QueryEndColumns,
    TraceEvents.QUERY_SUBCUBE: QuerySubcubeColumns,
    TraceEvents.SERVER_STATE_DISCOVER_BEGIN: ServerStateDiscoverBeginColumns,
    TraceEvents.SERVER_STATE_DISCOVER_DATA: ServerStateDiscoverDataColumns,
    TraceEvents.SERVER_STATE_DISCOVER_END: ServerStateDiscoverEndColumns,
    TraceEvents.SESSION_INITIALIZE: SessionInitializeColumns,
    TraceEvents.USER_DEFINED: UserDefinedColumns,
    TraceEvents.VERTIPAQ_SE_QUERY_BEGIN: VertipaqSeQueryBeginColumns,
    TraceEvents.VERTIPAQ_SE_QUERY_END: VertipaqSeQueryEndColumns,
    TraceEvents.VERTIPAQ_SE_QUERY_CACHE_MATCH: VertipaqSeQueryCacheMatchColumns,
    TraceEvents.VERTIPAQ_SE_QUERY_CACHE_MISS: VertipaqSeQueryCacheMissColumns,
    TraceEvents.WLGROUP_CPU_THROTTLING: WlgroupCpuThrottlingColumns,
    TraceEvents.WLGROUP_EXCEEDS_MEMORY_LIMIT: WlgroupExceedsMemoryLimitColumns,
    TraceEvents.WLGROUP_EXCEEDS_PROCESSING_LIMIT: WlgroupExceedsProcessingLimitColumns,
}



import datetime
from typing import TYPE_CHECKING, Literal

from pydantic import PrivateAttr

from pbi_core.lineage import LineageNode
from pbi_core.ssas.model_tables.base import SsasEditableRecord
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.measure import Measure


class KPI(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/1289ceca-8113-4019-8f90-8132a91117cf)
    """

    description: str | None = None
    measure_id: int
    status_description: str | None = None
    status_expression: str
    status_graphic: str
    target_description: str | None = None
    target_expression: str
    target_format_string: str
    trend_description: str | None = None
    trend_expression: str | None = None

    modified_time: datetime.datetime

    _commands: BaseCommands = PrivateAttr(default_factory=lambda: SsasCommands.kpi)

    def pbi_core_name(self) -> str:
        """Returns the name displayed in the PBIX report."""
        return self.measure().pbi_core_name()

    def measure(self) -> "Measure":
        return self.tabular_model.measures.find({"id": self.measure_id})

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(self, lineage_type)
        return LineageNode(self, lineage_type, [self.measure().get_lineage(lineage_type)])

    @classmethod
    def _db_command_obj_name(cls) -> str:
        return "Kpis"



from enum import IntEnum
from typing import Annotated, Any

from pydantic import Discriminator, Field, Tag

from pbi_core.static_files.layout._base_node import LayoutNode

from .column import ColumnSource
from .hierarchy import HierarchyLevelSource
from .measure import MeasureSource


class ExpressionName(LayoutNode):
    ExpressionName: str


class SelectRef(LayoutNode):
    SelectRef: ExpressionName


def get_expression_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "Aggregation" in v:
            return "AggregationSource"
        if any(c in v for c in ("Column", "Measure", "HierarchyLevel")):
            return "DataSource"
        if "SelectRef" in v:
            return "SelectRef"
        raise TypeError(v)
    return v.__class__.__name__


class AllRolesRef(LayoutNode):
    AllRolesRef: dict[str, bool] = Field(default_factory=dict)  # no values have been seen in this field


class ScopedEval2(LayoutNode):
    Expression: "ScopedEvalExpression"
    Scope: list[AllRolesRef]


class ScopedEval(LayoutNode):  # copied from arithmetic.py to avoid circular dependencies
    ScopedEval: ScopedEval2


def get_data_source_type(v: object | dict[str, Any]) -> str:
    if isinstance(v, dict):
        if "Column" in v:
            return "ColumnSource"
        if "Measure" in v:
            return "MeasureSource"
        if "HierarchyLevel" in v:
            return "HierarchyLevelSource"
        if "ScopedEval" in v:  # Consider subclassing? This only happens for color gradient properties IME
            return "ScopedEval"
        raise TypeError(v)
    return v.__class__.__name__


DataSource = Annotated[
    Annotated[ColumnSource, Tag("ColumnSource")]
    | Annotated[MeasureSource, Tag("MeasureSource")]
    | Annotated[HierarchyLevelSource, Tag("HierarchyLevelSource")]
    | Annotated[ScopedEval, Tag("ScopedEval")],
    Discriminator(get_data_source_type),
]


class AggregationFunction(IntEnum):
    SUM = 0
    AVERAGE = 1
    COUNT = 2
    MIN = 3
    MAX = 4
    DISTINCT_COUNT = 5
    MEDIAN = 6
    STD_DEV_P = 7
    VAR_P = 8


class _AggregationSourceHelper(LayoutNode):
    Expression: DataSource
    Function: AggregationFunction


class AggregationSource(LayoutNode):
    Aggregation: _AggregationSourceHelper
    Name: str | None = None
    NativeReferenceName: str | None = None  # only for Layout.Visual.Query

    def get_sources(self) -> list[DataSource]:
        return [self.Aggregation.Expression]


ScopedEvalExpression = Annotated[
    Annotated[DataSource, Tag("DataSource")]
    | Annotated[AggregationSource, Tag("AggregationSource")]
    | Annotated[SelectRef, Tag("SelectRef")],
    Discriminator(get_expression_type),
]



from pydantic import BaseModel, ConfigDict


class BaseValidation(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        use_enum_values=False,
        json_schema_mode_override="serialization",
        validate_assignment=True,
        protected_namespaces=(),
    )

    def __format__(self, format_spec: str) -> str:
        return f"{self!s:{format_spec}}"



import datetime
from typing import TYPE_CHECKING

from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasEditableRecord
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.column import Column
    from pbi_core.ssas.model_tables.perspective_table import PerspectiveTable


class PerspectiveColumn(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/f468353f-81a9-4a95-bb66-8997602bcd6d)
    """

    column_id: int
    perspective_table_id: int

    modified_time: datetime.datetime

    _commands: BaseCommands = PrivateAttr(default_factory=lambda: SsasCommands.perspective_column)

    def perspective_table(self) -> "PerspectiveTable":
        return self.tabular_model.perspective_tables.find(self.perspective_table_id)

    def column(self) -> "Column":
        return self.tabular_model.columns.find(self.column_id)





from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual
from .column_property import ColumnProperty
from .properties.base import Expression
from .text_box import GeneralProperties


class CategoryAxisProperties(LayoutNode):
    axisType: Expression | None = None
    concatenateLabels: Expression | None = None
    fontSize: Expression | None = None
    gridlineColor: Expression | None = None
    gridlineShow: Expression | None = None
    gridlineStyle: Expression | None = None
    innerPadding: Expression | None = None
    labelColor: Expression | None = None
    maxMarginFactor: Expression | None = None
    preferredCategoryWidth: Expression | None = None
    show: Expression | None = None
    showAxisTitle: Expression | None = None
    titleColor: Expression | None = None
    titleFontSize: Expression | None = None
    titleText: Expression | None = None


class CategoryAxis(LayoutNode):
    properties: CategoryAxisProperties
    selector: Selector | None = None


class DataPointProperties(LayoutNode):
    fill: Expression | None = None
    fillRule: Expression | None = None
    showAllDataPoints: Expression | None = None


class DataPoint(LayoutNode):
    properties: DataPointProperties
    selector: Selector | None = None


class LabelsProperties(LayoutNode):
    backgroundColor: Expression | None = None
    backgroundTransparency: Expression | None = None
    color: Expression | None = None
    enableBackground: Expression | None = None
    fontSize: Expression | None = None
    labelDensity: Expression | None = None
    labelDisplayUnits: Expression | None = None
    labelOrientation: Expression | None = None
    labelPosition: Expression | None = None
    show: Expression | None = None
    showAll: Expression | None = None


class Labels(LayoutNode):
    properties: LabelsProperties
    selector: Selector | None = None


class LegendProperties(LayoutNode):
    fontSize: Expression | None = None
    labelColor: Expression | None = None
    position: Expression | None = None
    show: Expression | None = None
    showTitle: Expression | None = None


class Legend(LayoutNode):
    properties: LegendProperties
    selector: Selector | None = None


class ValueAxisProperties(LayoutNode):
    axisScale: Expression | None = None
    fontSize: Expression | None = None
    gridlineShow: Expression | None = None
    logAxisScale: Expression | None = None
    show: Expression | None = None
    showAxisTitle: Expression | None = None
    start: Expression | None = None
    titleFontFamily: Expression | None = None


class ValueAxis(LayoutNode):
    properties: ValueAxisProperties
    selector: Selector | None = None


class General(LayoutNode):
    properties: GeneralProperties


class ZoomProperties(LayoutNode):
    show: Expression | None = None


class Zoom(LayoutNode):
    properties: ZoomProperties


class TotalProperties(LayoutNode):
    show: Expression | None = None


class Total(LayoutNode):
    properties: TotalProperties


class Y1AxisReferenceLineProperties(LayoutNode):
    displayName: Expression | None = None
    lineColor: Expression | None = None
    show: Expression | None = None
    style: Expression | None = None
    transparency: Expression | None = None
    value: Expression | None = None


class Y1AxisReferenceLine(LayoutNode):
    properties: Y1AxisReferenceLineProperties
    selector: Selector | None = None


class ColumnChartColumnProperties(LayoutNode):
    categoryAxis: list[CategoryAxis] | None = None
    dataPoint: list[DataPoint] | None = None
    general: list[General] | None = None
    labels: list[Labels] | None = None
    legend: list[Legend] | None = None
    valueAxis: list[ValueAxis] | None = None
    totals: list[Total] | None = None
    y1AxisReferenceLine: list[Y1AxisReferenceLine] | None = None
    zoom: list[Zoom] | None = None


class ColumnChart(BaseVisual):
    visualType: str = "columnChart"
    model_config = ConfigDict(extra="forbid")

    objects: ColumnChartColumnProperties | None = None
    columnProperties: dict[str, ColumnProperty] | None = None



import datetime
from typing import TYPE_CHECKING

from pydantic import PrivateAttr

from pbi_core.ssas.model_tables.base import SsasEditableRecord
from pbi_core.ssas.server._commands import BaseCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.hierarchy import Hierarchy
    from pbi_core.ssas.model_tables.perspective_table import PerspectiveTable


class PerspectiveHierarchy(SsasEditableRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/07941935-98bf-4e14-ab40-ef97d5c29765)
    """

    hierarchy_id: int
    perspective_table_id: int

    modified_time: datetime.datetime

    _commands: BaseCommands = PrivateAttr(default_factory=lambda: SsasCommands.perspective_hierarchy)

    def perspective_table(self) -> "PerspectiveTable":
        return self.tabular_model.perspective_tables.find(self.perspective_table_id)

    def hierarchy(self) -> "Hierarchy":
        return self.tabular_model.hierarchies.find(self.hierarchy_id)



import pydantic
from bs4 import BeautifulSoup
from structlog import get_logger

from pbi_core.ssas.server._commands import BaseCommands, ModelCommands, NoCommands, RefreshCommands, RenameCommands

from .base_ssas_table import SsasTable
from .enums import RefreshType

logger = get_logger()


SsasConfig = pydantic.ConfigDict(
    arbitrary_types_allowed=True,
    extra="forbid",
    use_enum_values=False,
    json_schema_mode_override="serialization",
    validate_assignment=True,
    protected_namespaces=(),
)


logger = get_logger()


class SsasAlter(SsasTable):
    """Class for SSAS records that implement alter functionality.

    The `alter <https://learn.microsoft.com/en-us/analysis-services/tmsl/alter-command-tmsl?view=asallproducts-allversions>`_ spec
    """  # noqa: E501

    _commands: BaseCommands

    def alter(self) -> BeautifulSoup:
        """Updates a non-name field of an object."""
        xml_command = self.render_xml_command(
            self.xml_fields(),
            self._commands.alter,
            self.tabular_model.db_name,
        )
        logger.info("Syncing Alter Changes to SSAS", obj=self._db_type_name())
        return self.query_xml(xml_command, db_name=self.tabular_model.db_name)


class SsasRename(SsasTable):
    """Class for SSAS records that implement rename functionality.

    The `rename <https://learn.microsoft.com/en-us/analysis-services/tmsl/rename-command-tmsl?view=asallproducts-allversions>`_ spec
    """  # noqa: E501

    _db_name_field: str = "not_defined"
    _commands: RenameCommands

    def rename(self) -> BeautifulSoup:
        """Updates a name field of an object."""
        xml_command = self.render_xml_command(
            self.xml_fields(),
            self._commands.rename,
            self.tabular_model.db_name,
        )
        logger.info("Syncing Rename Changes to SSAS", obj=self._db_type_name())
        return self.query_xml(xml_command, db_name=self.tabular_model.db_name)


class SsasCreate(SsasTable):
    """Class for SSAS records that implement create functionality.

    The `create <https://learn.microsoft.com/en-us/analysis-services/tmsl/create-command-tmsl?view=asallproducts-allversions>`_ spec
    """  # noqa: E501

    _commands: BaseCommands

    def create(self) -> BeautifulSoup:
        """Creates a new SSAS object based on the python object."""
        xml_command = self.render_xml_command(
            self.xml_fields(),
            self._commands.create,
            self.tabular_model.db_name,
        )
        logger.info("Syncing Create Changes to SSAS", obj=self._db_type_name())
        return self.tabular_model.server.query_xml(xml_command, db_name=self.tabular_model.db_name)


class SsasDelete(SsasTable):
    """Class for SSAS records that implement delete functionality.

    The `delete <https://learn.microsoft.com/en-us/analysis-services/tmsl/delete-command-tmsl?view=asallproducts-allversions>`_ spec
    """  # noqa: E501

    _db_id_field: str = "id"  # we're comparing the name before the translation back to SSAS casing
    _commands: BaseCommands

    def delete(self) -> BeautifulSoup:
        data = {
            self._db_id_field: getattr(self, self._db_id_field),
        }
        xml_command = self.render_xml_command(
            data,
            self._commands.delete,
            self.tabular_model.db_name,
        )
        logger.info("Syncing Delete Changes to SSAS", obj=self._db_type_name())
        return self.query_xml(xml_command, db_name=self.tabular_model.db_name)


class SsasRefresh(SsasTable):
    """Class for SSAS records that implement refresh functionality.

    The `refresh <https://learn.microsoft.com/en-us/analysis-services/tmsl/refresh-command-tmsl?view=asallproducts-allversions>`_ spec
    """  # noqa: E501

    _db_id_field: str = "id"  # we're comparing the name before the translation back to SSAS casing
    _default_refresh_type: RefreshType
    _commands: RefreshCommands

    def refresh(self, refresh_type: RefreshType | None = None) -> BeautifulSoup:
        xml_command = self.render_xml_command(
            self.xml_fields(),
            self._commands.refresh,
            self.tabular_model.db_name,
        )
        logger.info("Syncing Refresh Changes to SSAS", obj=self)
        return self.query_xml(xml_command, db_name=self.tabular_model.db_name)


class SsasReadonlyRecord(SsasTable):
    """Class for SSAS records that implement no command."""

    _commands: NoCommands


class SsasEditableRecord(SsasCreate, SsasAlter, SsasDelete):
    _commands: BaseCommands


class SsasRenameRecord(SsasCreate, SsasAlter, SsasDelete, SsasRename):
    _commands: RenameCommands  # pyright: ignore reportIncompatibleVariableOverride


class SsasRefreshRecord(SsasCreate, SsasAlter, SsasDelete, SsasRename, SsasRefresh):
    _commands: RefreshCommands  # pyright: ignore reportIncompatibleVariableOverride


class SsasModelRecord(SsasAlter, SsasRefresh, SsasRename):
    """Solely used for the single Model record."""

    _commands: ModelCommands  # pyright: ignore reportIncompatibleVariableOverride



import datetime
from typing import TYPE_CHECKING, Literal

from pydantic import Json, PrivateAttr

from pbi_core.lineage import LineageNode
from pbi_core.pydantic.main import BaseValidation
from pbi_core.ssas.model_tables.base import RefreshType, SsasModelRecord
from pbi_core.ssas.server._commands import ModelCommands
from pbi_core.ssas.server.utils import SsasCommands

from .enums import DefaultDataView

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.culture import Culture
    from pbi_core.ssas.model_tables.measure import Measure
    from pbi_core.ssas.model_tables.query_group import QueryGroup
    from pbi_core.ssas.model_tables.table import Table


class DataAccessOptions(BaseValidation):
    fastCombine: bool = True
    legacyRedirects: bool = False
    returnErrorValuesAsNull: bool = False


class Model(SsasModelRecord):
    """tbd.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/60094cd5-1c7e-4353-9299-251bfa838cc6)
    """

    _default_refresh_type = RefreshType.CALCULATE

    automatic_aggregation_options: str | None = None
    collation: str | None = None
    culture: str
    data_access_options: Json[DataAccessOptions] = DataAccessOptions()
    data_source_default_max_connections: int
    data_source_variables_override_behavior: int
    default_data_view: DefaultDataView
    default_measure_id: int | None = None
    default_mode: int
    default_powerbi_data_source_version: int
    description: str | None = None
    discourage_composite_models: bool = True
    discourage_implicit_measures: bool
    disable_auto_exists: int | None = None
    force_unique_names: bool
    m_attributes: str | None = None
    max_parallelism_per_refresh: int | None = None
    max_parallelism_per_query: int | None = None
    name: str
    source_query_culture: str = "en-US"
    storage_location: str | None = None
    version: int

    modified_time: datetime.datetime
    structure_modified_time: datetime.datetime

    _commands: ModelCommands = PrivateAttr(default_factory=lambda: SsasCommands.model)

    def default_measure(self) -> "Measure | None":
        if self.default_measure_id is None:
            return None
        return self.tabular_model.measures.find(self.default_measure_id)

    def cultures(self) -> set["Culture"]:
        return self.tabular_model.cultures.find_all({"model_id": self.id})

    def tables(self) -> set["Table"]:
        return self.tabular_model.tables.find_all({"model_id": self.id})

    def query_groups(self) -> set["QueryGroup"]:
        return self.tabular_model.query_groups.find_all({"model_id": self.id})

    @classmethod
    def _db_command_obj_name(cls) -> str:
        return "Model"

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(
                self,
                lineage_type,
                [c.get_lineage(lineage_type) for c in self.cultures()]
                + [t.get_lineage(lineage_type) for t in self.tables()]
                + [q.get_lineage(lineage_type) for q in self.query_groups()],
            )
        return LineageNode(self, lineage_type)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.attribute_hierarchy import AttributeHierarchy
    from pbi_core.ssas.model_tables.level import Level
    from pbi_core.ssas.model_tables.relationship.relationship import Relationship
    from pbi_core.ssas.model_tables.table import Table
    from pbi_core.ssas.server.tabular_model import BaseTabularModel

    from .column import Column


class RelationshipMixin:
    id: int
    table_id: int
    attribute_hierarchy_id: int
    sort_by_column_id: int | None
    tabular_model: "BaseTabularModel"

    def table(self) -> "Table":
        """Returns the table class the column is a part of."""
        return self.tabular_model.tables.find({"id": self.table_id})

    def attribute_hierarchy(self) -> "AttributeHierarchy":
        return self.tabular_model.attribute_hierarchies.find({"id": self.attribute_hierarchy_id})

    def levels(self) -> set["Level"]:
        return self.tabular_model.levels.find_all({"column_id": self.id})

    def sort_by_column(self) -> "Column | None":
        if self.sort_by_column_id is None:
            return None
        return self.tabular_model.columns.find({"id": self.sort_by_column_id})

    def sorting_columns(self) -> set["Column"]:
        """Provides the inverse information of sort_by_column."""
        return self.tabular_model.columns.find_all({"sort_by_column_id": self.id})

    def from_relationships(self) -> set["Relationship"]:
        return self.tabular_model.relationships.find_all({"from_column_id": self.id})

    def to_relationships(self) -> set["Relationship"]:
        return self.tabular_model.relationships.find_all({"to_column_id": self.id})

    def relationships(self) -> set["Relationship"]:
        return self.from_relationships() | self.to_relationships()



from pydantic import ConfigDict

from pbi_core.static_files.layout._base_node import LayoutNode
from pbi_core.static_files.layout.selector import Selector

from .base import BaseVisual
from .column_property import ColumnProperty
from .properties.base import Expression


class LabelsPropertiesHelper(LayoutNode):
    backgroundTransparency: Expression | None = None
    color: Expression | None = None
    enableBackground: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    labelDisplayUnits: Expression | None = None
    labelOrientation: Expression | None = None
    labelOverflow: Expression | None = None
    labelPosition: Expression | None = None
    labelPrecision: Expression | None = None
    show: Expression | None = None
    showAll: Expression | None = None


class LabelsProperties(LayoutNode):
    properties: LabelsPropertiesHelper
    selector: Selector | None = None


class ValueAxisPropertiesHelper(LayoutNode):
    axisScale: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    gridlineShow: Expression | None = None
    labelColor: Expression | None = None
    labelDisplayUnits: Expression | None = None
    logAxisScale: Expression | None = None
    show: Expression | None = None
    showAxisTitle: Expression | None = None
    start: Expression | None = None
    titleFontFamily: Expression | None = None


class ValueAxisProperties(LayoutNode):
    properties: ValueAxisPropertiesHelper
    selector: Selector | None = None


class DataPointPropertiesHelper(LayoutNode):
    fill: Expression | None = None
    showAllDataPoints: Expression | None = None


class DataPointProperties(LayoutNode):
    properties: DataPointPropertiesHelper
    selector: Selector | None = None


class CategoryAxisPropertiesHelper(LayoutNode):
    axisType: Expression | None = None
    concatenateLabels: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    gridlineShow: Expression | None = None
    gridlineStyle: Expression | None = None
    innerPadding: Expression | None = None
    labelColor: Expression | None = None
    maxMarginFactor: Expression | None = None
    preferredCategoryWidth: Expression | None = None
    show: Expression | None = None
    showAxisTitle: Expression | None = None
    titleColor: Expression | None = None
    titleFontSize: Expression | None = None


class CategoryAxisProperties(LayoutNode):
    properties: CategoryAxisPropertiesHelper


class LegendPropertiesHelper(LayoutNode):
    fontSize: Expression | None = None
    labelColor: Expression | None = None
    position: Expression | None = None
    show: Expression | None = None
    showTitle: Expression | None = None


class LegendProperties(LayoutNode):
    properties: LegendPropertiesHelper


class TrendPropertiesHelper(LayoutNode):
    displayName: Expression | None = None
    lineColor: Expression | None = None
    show: Expression | None = None


class TrendProperties(LayoutNode):
    properties: TrendPropertiesHelper


class PlotAreaPropertiesHelper(LayoutNode):
    transparency: Expression | None = None


class PlotAreaProperties(LayoutNode):
    properties: PlotAreaPropertiesHelper


class GeneralPropertiesHelper(LayoutNode):
    responsive: Expression | None = None


class GeneralProperties(LayoutNode):
    properties: GeneralPropertiesHelper


class ClusteredColumnChartProperties(LayoutNode):
    categoryAxis: list[CategoryAxisProperties] | None = None
    dataPoint: list[DataPointProperties] | None = None
    general: list[GeneralProperties] | None = None
    labels: list[LabelsProperties] | None = None
    legend: list[LegendProperties] | None = None
    plotArea: list[PlotAreaProperties] | None = None
    trend: list[TrendProperties] | None = None
    valueAxis: list[ValueAxisProperties] | None = None


class ClusteredColumnChart(BaseVisual):
    visualType: str = "clusteredColumnChart"
    model_config = ConfigDict(extra="forbid")

    columnProperties: dict[str, ColumnProperty] | None = None
    drillFilterOtherVisuals: bool = True
    objects: ClusteredColumnChartProperties | None = None



import pathlib
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any

import backoff
import pbi_pyadomd
import psutil
from bs4 import BeautifulSoup

from pbi_core.logging import get_logger
from pbi_core.ssas.setup import get_startup_config

from ._physical_local_server import SSASProcess
from .tabular_model import BaseTabularModel, LocalTabularModel
from .utils import COMMAND_TEMPLATES, SKU_ERROR, get_msmdsrv_info

logger = get_logger()

if TYPE_CHECKING:
    from _typeshed import StrPath


DT_FORMAT = "%Y-%m-%d_%H-%M-%S"


class BaseServer:
    """Base Server Interface containing the methods used outside of instance lifetime management."""

    host: str
    "A hostname to the background SSAS instance"

    port: int
    "A Port to the background SSAS instance"

    default_db: str | None
    """The default DB to use when executing DAX"""

    def __init__(self, host: str, port: int, default_db: str | None = None) -> None:
        self.host = host
        self.port = port
        self.default_db = default_db

        self.check_ssas_sku()

    def conn_str(self, db_name: str | None = None) -> str:
        """Formats the connection string for connecting to the background SSAS instance."""
        if db_name:
            return f"Provider=MSOLAP;Data Source={self.host}:{self.port};Initial Catalog={db_name};"
        return f"Provider=MSOLAP;Data Source={self.host}:{self.port};"

    def conn(self, db_name: str | None = None) -> pbi_pyadomd.Connection:
        """Returns a pbi_pyadomd connection."""
        return pbi_pyadomd.connect(self.conn_str(db_name))

    def __repr__(self) -> str:
        return f"Server(host={self.host}:{self.port})"

    def query_dax(self, query: str, *, db_name: str | bool | None = True) -> list[dict[str, Any]]:
        """db_name: when bool and == True, uses the DB last loaded by this server instance.

        (almost always the db of the loaded PBI unless you're manually reassigning server instances)
        when None or False, no db is supplied
        when a string, just passed to the client
        """
        if db_name is True:
            db_name = self.default_db
        elif db_name is False:
            db_name = None
        with self.conn(db_name) as conn:
            reader = conn.execute_dax(query)
            return reader.fetch_many()

    def query_xml(self, query: str, db_name: str | None = None) -> BeautifulSoup:
        """Submits an XMLA query to the SSAS instance and returns the result as a BeautifulSoup object.

        The query should be a valid XMLA command.

        Args:
            query (str): The XMLA query to execute.
            db_name (str | None): The name of the database to execute the query against.

        Returns:
            BeautifulSoup: The result of the query parsed as a BeautifulSoup object.

        """
        with self.conn(db_name) as conn:
            return conn.execute_xml(query)

    def tabular_models(self) -> list[BaseTabularModel]:
        """Creates a list of the Tabular models existing in the SSAS server.

        Note:
            Generally tabular models in the local environment correspond 1-1 with a PBIX report open in the Desktop app

        """
        # Query based on https://learn.microsoft.com/en-us/previous-versions/sql/sql-server-2012/ms126314(v=sql.110)
        dbs = self.query_dax(COMMAND_TEMPLATES["list_dbs.xml"].render())
        return [BaseTabularModel(row["CATALOG_NAME"], self) for row in dbs]

    @backoff.on_exception(backoff.expo, ValueError, max_time=10)
    def check_ssas_sku(self) -> None:
        """Checks if the SSAS instance is running the correct SKU version.

        Tests this assumption by running a query that should fail if the SKU under 1400 (image save command).
        Since we could also fail due to the server not being instantiated, we use a backoff decorator
        to retry the command a few times before giving up.

        Raises:
            TypeError: If the SSAS instance is running an incorrect SKU version.
            ValueError: If the SSAS instance is not running or the command fails for another reason.
                This is used to trigger the backoff retry.

        """
        try:
            self.query_xml(
                COMMAND_TEMPLATES["image_save.xml"].render(
                    target_path="---",
                    db_name="---",
                ),  # specifically choosing non-existant values to verify we get at least one error
            )
        except pbi_pyadomd.AdomdErrorResponseException as e:
            error_message = str(e.Message)
            if error_message == SKU_ERROR:
                return
            msg = f"Incorrect SKUVersion. We got the error: {error_message}"
            raise TypeError(msg) from None
        msg = "Got a 'file not loaded' type error. Waiting"
        raise ValueError(msg)

    @staticmethod
    def sanitize_xml(xml_text: str) -> str:
        """Method to XML-encode characters like "&" so that the Adomd connection doesn't mangle the XMLA commands."""
        return xml_text.replace("&", "&amp;")

    @staticmethod
    def remove_invalid_db_name_chars(orig_db_name: str) -> str:
        """Utility function to convert a PBIX report name to an equivalent name for the DB in the SSAS instance.

        Note:
            Raises a warning if the db_name is changed to inform user that the db_name does not match their input

        """
        db_name = orig_db_name.replace("&", " ")[:100]
        db_name = db_name.strip()  # needed to find the correct name, since SSAS does stripping too
        if orig_db_name != db_name:
            logger.warning("db_name changed", original_name=orig_db_name, new_name=db_name)
        return db_name


class LocalServer(BaseServer):
    """A Server running locally on the user's machine.

    This subclass has the ability to load/dump Tabular Models
    to a PBIX file. Also creates a background SSAS instance and workspace to handle processing if none is provided.

    Args:
        kill_on_exit (bool): Indicates if the background SSAS instance handling
            processing should be terminated at the end of the python session

    """

    physical_process: SSASProcess
    """
    A Python class handling the lifetime of the SSAS Instance. Interacts with the SSAS instance only as a process
    """

    def __init__(
        self,
        host: str = "localhost",
        workspace_directory: "StrPath | None" = None,
        pid: int | None = None,
        *,
        kill_on_exit: bool = True,
    ) -> None:
        if pid is not None:
            self.physical_process = SSASProcess(pid=pid, kill_on_exit=kill_on_exit, startup_config=None)
        else:
            startup_config = get_startup_config()
            workspace_directory = workspace_directory or (
                startup_config.workspace_dir / datetime.now(UTC).strftime(DT_FORMAT)
            )
            self.physical_process = SSASProcess(
                workspace_directory=workspace_directory,
                kill_on_exit=kill_on_exit,
                startup_config=startup_config,
            )
        super().__init__(host, self.physical_process._port)

    def load_pbix(self, path: "StrPath", *, db_name: str | None = None) -> LocalTabularModel:
        """Takes a Path to a PBIX report and loads the PBIX Datamodel to the SSAS instance in the SSASProcess.

        Raises:
            FileNotFoundError: when the path to the PBIX file does not exist
            AdomdErrorResponseException: Occurs when the DB already exists

        """
        path = pathlib.Path(path)
        if not path.exists():
            msg = f"The path to the PBIX does not exist: {path.absolute().as_posix()}"
            raise FileNotFoundError(msg)
        if db_name is None:
            db_name = path.stem
        db_name = self.remove_invalid_db_name_chars(db_name)
        load_command = COMMAND_TEMPLATES["image_load.xml"].render(
            db_name=db_name,
            source_path=self.sanitize_xml(path.absolute().as_posix()),
        )
        try:
            self.query_xml(load_command)
        except pbi_pyadomd.AdomdErrorResponseException as e:
            if (
                "user does not have permission to restore the database, or the database already exists and AllowOverwrite is not specified"  # noqa: E501
                in str(e.Message)
            ):
                logger.warning("Removing old version of PBIX data model for new version", db_name=db_name)
                self.query_xml(COMMAND_TEMPLATES["db_delete.xml"].render(db_name=db_name))
                self.query_xml(load_command)
            else:
                raise pbi_pyadomd.AdomdErrorResponseException from e

        self.default_db = db_name  # needed so the DAX queries are pointed to the right DB by defauilt
        logger.info("Tabular Model load complete")
        tab_model = LocalTabularModel(db_name=db_name, server=self, pbix_path=path)
        tab_model.sync_from()
        return tab_model

    def save_pbix(self, path: "StrPath", db_name: str) -> None:
        path = pathlib.Path(path)
        self.query_xml(
            COMMAND_TEMPLATES["image_save.xml"].render(
                db_name=db_name,
                target_path=self.sanitize_xml(path.absolute().as_posix()),
            ),
        )

    def __repr__(self) -> str:
        return f"LocalServer(port={self.port})"


def list_local_servers() -> list[LocalServer]:
    """Returns all active SSAS instances on a computer accessible from the python instance.

    Note:
        The main thing that would block a SSAS instance from being verified by a python
        instance is insufficient permissions

    """
    ret: list[LocalServer] = [
        LocalServer(pid=process.pid, kill_on_exit=False)
        for process in psutil.process_iter()
        if get_msmdsrv_info(process) is not None
    ]
    return ret


def get_or_create_local_server(*, kill_on_exit: bool = True) -> LocalServer:
    """Gets a local server to load the PBIX to.

    Checks the list of active processes on your local machine for a ``msmdsrv.exe`` process with an active port and
    a corresponding workspace folder. If no matching process is found, this function generates a new process.

    Args:
        kill_on_exit (bool, optional): **If** this function creates a new instance of a local SSAS process, this
            argument will control if the process is killed at the end of the Python session.

    """
    candidates: list[LocalServer] = list_local_servers()
    if candidates:
        logger.info("Re-using existing local SSAS instance", port=candidates[0].port)
        return candidates[0]
    return LocalServer(kill_on_exit=kill_on_exit)


def terminate_all_local_servers() -> None:
    """Attempts to terminate all SSAS instances on a computer.

    Useful for cleaning up the environment and avoiding memory leaks
    """
    for server in list_local_servers():
        server.physical_process.terminate()



from enum import IntEnum


class RelationshipType(IntEnum):
    SINGLE_COLUMN = 1


class CrossFilteringBehavior(IntEnum):
    ONE_DIRECTION = 1
    BOTH_DIRECTION = 2
    AUTOMATIC = 3


class JoinOnDateBehavior(IntEnum):
    DATE_AND_TIME = 1
    DATE_PART_ONLY = 2


class SecurityFilteringBehavior(IntEnum):
    ONE_DIRECTION = 1
    BOTH_DIRECTIONS = 2
    _NONE = 3



import time
from pathlib import Path

import bs4
import requests
from structlog import get_logger

logger = get_logger()

BASE_URL = "https://learn.microsoft.com/en-us/analysis-services/trace-events"
SEP = "\n" * 3


def gen_trace_enums(class_name: str, url: str) -> str:
    logger.info("Processing Events")
    ret = [f"class {class_name}(IntEnum):"]
    page = requests.get(BASE_URL + url, timeout=10)
    content = bs4.BeautifulSoup(page.content, features="lxml")
    for table in content.find_all("table"):
        for row in list(table.find_all("tr"))[1:]:  # pyright: ignore reportAttributeAccessIssue
            children = list(row.find_all("td"))  # pyright: ignore reportAttributeAccessIssue
            ret.append(
                f"\t{str(children[1].text).replace(' ', '_').upper()} = {children[0].text}  # {children[2].text}",
            )
    return "\n".join(ret) + SEP


def gen_event_enums(url: str) -> str:
    logger.info("Processing Columns", url=url)
    time.sleep(2)  # needed to avoid rate limiting
    ret: list[str] = []
    page = requests.get(BASE_URL + url, timeout=10)
    content = bs4.BeautifulSoup(page.content, features="lxml")
    headers: list[str] = [x.text for x in content.find_all("h2")][1:-3]
    headers = [x[: (x + " Class—Data").index(" Class—Data")].replace(" ", "") + "Columns" for x in headers]
    tables = content.find_all("table")
    tables = tables[len(tables) - len(headers) :]
    for header, table in zip(headers, tables, strict=False):
        ret2 = [f"class {header}(IntEnum):"]
        if table.find("tr").find("th").text != "Column Name":  # pyright: ignore reportAttributeAccessIssue
            continue
        for row in list(table.find_all("tr"))[1:]:  # pyright: ignore reportAttributeAccessIssue
            children = list(row.find_all("td"))  # pyright: ignore reportAttributeAccessIssue
            comment = children[len(children) - 1].text.replace("\n", " ")
            ret2.append(f"\t{str(children[0].text).replace(' ', '_').upper()} = {children[1].text}  # {comment}")
        ret.append("\n".join(ret2) + SEP)
    return "\n".join(ret)


with (Path(__file__).parent / "trace_enums.py").open("w", encoding="utf-8") as f:
    f.write("from enum import IntEnum" + SEP)
    f.write(gen_trace_enums("TraceEvents", "/analysis-services-trace-events"))
    # f.write(gen_event_enums("/command-events-data-columns"))
    f.write(gen_event_enums("/discover-events-data-columns"))
    f.write(gen_event_enums("/discover-server-state-events-data-columns"))
    f.write(gen_event_enums("/errors-and-warnings-events-data-columns"))
    f.write(gen_event_enums("/file-load-and-save-data-columns"))
    f.write(gen_event_enums("/lock-events-data-columns"))
    # f.write(gen_event_enums("/progress-reports-data-columns"))
    f.write(gen_event_enums("/security-audit-data-columns"))
    # f.write(gen_event_enums("/jobgraph-events-data-columns"))
    f.write(gen_event_enums("/notification-events-data-columns"))
    f.write(gen_event_enums("/queries-events-data-columns"))
    f.write(gen_event_enums("/query-processing-events-data-column"))
    f.write(gen_event_enums("/session-events-data-columns"))



from pbi_core.static_files.layout._base_node import LayoutNode

from .base import Expression


class BackgroundProperties(LayoutNode):
    color: Expression | None = None
    show: Expression | None = None
    transparency: Expression | None = None


class Background(LayoutNode):
    properties: BackgroundProperties


class GeneralProperties(LayoutNode):
    altText: Expression | None = None
    keepLayerOrder: Expression | None = None


class General(LayoutNode):
    properties: GeneralProperties


class TitleProperties(LayoutNode):
    alignment: Expression | None = None
    background: Expression | None = None
    fontColor: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    heading: Expression | None = None
    show: Expression | None = None
    text: Expression | None = None
    titleWrap: Expression | None = None


class Title(LayoutNode):
    properties: TitleProperties


class VisualTooltipProperties(LayoutNode):
    background: Expression | None = None
    fontFamily: Expression | None = None
    fontSize: Expression | None = None
    section: Expression | None = None
    show: Expression | None = None
    titleFontColor: Expression | None = None
    type: Expression | None = None
    valueFontColor: Expression | None = None


class VisualTooltip(LayoutNode):
    properties: VisualTooltipProperties


class StylePresetProperties(LayoutNode):
    name: Expression | None = None


class StylePreset(LayoutNode):
    properties: StylePresetProperties


class BorderProperties(LayoutNode):
    background: Expression | None = None
    color: Expression | None = None
    radius: Expression | None = None
    show: Expression | None = None


class Border(LayoutNode):
    properties: BorderProperties


class VisualHeaderProperties(LayoutNode):
    background: Expression | None = None
    border: Expression | None = None
    foreground: Expression | None = None
    show: Expression | None = None
    showDrillDownExpandButton: Expression | None = None
    showDrillDownLevelButton: Expression | None = None
    showDrillRoleSelector: Expression | None = None
    showDrillToggleButton: Expression | None = None
    showDrillUpButton: Expression | None = None
    showFilterRestatementButton: Expression | None = None
    showFocusModeButton: Expression | None = None
    showOptionsMenu: Expression | None = None
    showPinButton: Expression | None = None
    showSeeDataLayoutToggleButton: Expression | None = None
    showSmartNarrativeButton: Expression | None = None
    showTooltipButton: Expression | None = None
    showVisualErrorButton: Expression | None = None
    showVisualInformationButton: Expression | None = None
    showVisualWarningButton: Expression | None = None
    transparency: Expression | None = None


class VisualHeader(LayoutNode):
    properties: VisualHeaderProperties


class VisualLinkProperties(LayoutNode):
    bookmark: Expression | None = None
    disabledTooltip: Expression | None = None
    drillthroughSection: Expression | None = None
    enabledTooltip: Expression | None = None
    navigationSection: Expression | None = None
    show: Expression | None = None
    tooltip: Expression | None = None
    type: Expression | None = None
    webUrl: Expression | None = None


class VisualLink(LayoutNode):
    properties: VisualLinkProperties | None = None


class VisualHeaderTooltipProperties(LayoutNode):
    background: Expression | None = None
    section: Expression | None = None
    text: Expression | None = None
    titleFontColor: Expression | None = None
    type: Expression | None = None


class VisualHeaderTooltip(LayoutNode):
    properties: VisualHeaderTooltipProperties


class DropShadowProperties(LayoutNode):
    angle: Expression | None = None
    color: Expression | None = None
    position: Expression | None = None
    preset: Expression | None = None
    shadowBlur: Expression | None = None
    shadowDistance: Expression | None = None
    shadowSpread: Expression | None = None
    show: Expression | None = None
    transparency: Expression | None = None


class DropShadow(LayoutNode):
    properties: DropShadowProperties


class LockAspectProperties(LayoutNode):
    show: Expression | None = None


class LockAspect(LayoutNode):
    properties: LockAspectProperties


class VCProperties(LayoutNode):
    background: list[Background] | None = None
    border: list[Border] | None = None
    dropShadow: list[DropShadow] | None = None
    general: list[General] | None = None
    lockAspect: list[LockAspect] | None = None
    stylePreset: list[StylePreset] | None = None
    title: list[Title] | None = None
    visualHeader: list[VisualHeader] | None = None
    visualHeaderTooltip: list[VisualHeaderTooltip] | None = None
    visualLink: list[VisualLink] | None = None
    visualTooltip: list[VisualTooltip] | None = None



from .layout import (
    Layout,
    LayoutNode,
    ResourcePackage,
    ResourcePackageDetails,
    ResourcePackageDetailsType,
    ResourcePackageItem,
    ResourcePackageItemType,
    Section,
)
from .static_files import StaticFiles

__all__ = [
    "Layout",
    "LayoutNode",
    "ResourcePackage",
    "ResourcePackageDetails",
    "ResourcePackageDetailsType",
    "ResourcePackageItem",
    "ResourcePackageItemType",
    "Section",
    "StaticFiles",
]



from datetime import UTC, datetime

from pbi_core.static_files.layout._base_node import LayoutNode

PrimitiveValue = int | str | datetime | bool | None
DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"


def parse_literal(literal_val: str) -> PrimitiveValue:
    if literal_val == "null":
        return None
    if literal_val in {"true", "false"}:
        return literal_val == "true"
    if literal_val.endswith("L"):
        return int(literal_val[:-1])
    if literal_val.startswith("datetime"):
        return datetime.strptime(literal_val[9:-1], DATETIME_FORMAT).replace(tzinfo=UTC)
    return literal_val[1:-1]


class _LiteralSourceHelper(LayoutNode):
    Value: str


class LiteralSource(LayoutNode):
    Literal: _LiteralSourceHelper

    def value(self) -> PrimitiveValue:
        return parse_literal(self.Literal.Value)

    def __repr__(self) -> str:
        return f"LiteralSource({self.Literal.Value})"



import datetime
from typing import TYPE_CHECKING, Literal
from uuid import UUID, uuid4

from bs4 import BeautifulSoup
from pydantic import PrivateAttr

from pbi_core.lineage import LineageNode
from pbi_core.ssas.model_tables.base import RefreshType, SsasRefreshRecord
from pbi_core.ssas.model_tables.column import Column
from pbi_core.ssas.model_tables.enums import DataCategory
from pbi_core.ssas.model_tables.measure import Measure
from pbi_core.ssas.model_tables.partition import Partition
from pbi_core.ssas.server._commands import RefreshCommands
from pbi_core.ssas.server.utils import SsasCommands

if TYPE_CHECKING:
    from pbi_core.ssas.model_tables.calculation_group import CalculationGroup
    from pbi_core.ssas.model_tables.detail_row_definition import DetailRowDefinition
    from pbi_core.ssas.model_tables.model import Model
    from pbi_core.ssas.model_tables.refresh_policy import RefreshPolicy


class Table(SsasRefreshRecord):
    """TBD.

    SSAS spec: [Microsoft](https://learn.microsoft.com/en-us/openspecs/sql_server_protocols/ms-ssas-t/6360ac84-0717-4170-bce0-284cbef419ca)
    """

    _refresh_type = RefreshType.DATA_ONLY

    alternate_source_precedence: int
    calculation_group_id: int | None = None
    data_category: DataCategory | None = None
    default_detail_rows_defintion_id: int | None = None
    description: str | None = None
    """A description of the table, which may be used in the hover tooltip in edit mode"""
    exclude_from_automatic_aggregations: bool = False
    exclude_from_model_refresh: bool
    """Controls whether this table is included in the model-wide refresh process"""
    is_hidden: bool
    """Controls whether the table appears in the edit mode of the report"""
    is_private: bool
    model_id: int
    """The ID of the model this table belongs to"""
    name: str
    """The name of the table as it appears in the report"""
    refresh_policy_id: int | None = None
    show_as_variations_only: bool
    system_flags: int
    system_managed: bool | None = None
    table_storage_id: int | None = None

    lineage_tag: UUID = uuid4()
    source_lineage_tag: UUID = uuid4()

    modified_time: datetime.datetime
    structure_modified_time: datetime.datetime

    _commands: RefreshCommands = PrivateAttr(default_factory=lambda: SsasCommands.table)

    def modification_hash(self) -> int:
        return hash((
            self.name,
            self.description,
            self.is_hidden,
        ))

    def calculation_group(self) -> "CalculationGroup | None":
        if self.calculation_group_id is None:
            return None
        return self.tabular_model.calculation_groups.find(self.calculation_group_id)

    def refresh_policy(self) -> "RefreshPolicy | None":
        if self.refresh_policy_id is None:
            return None
        return self.tabular_model.refresh_policies.find(self.refresh_policy_id)

    def is_system_table(self) -> bool:
        return bool(self.system_flags >> 1 % 2)

    def is_from_calculated_table(self) -> bool:
        return bool(self.system_flags % 2)

    def data(self, head: int = 100) -> list[dict[str, int | float | str]]:
        """Extracts records from the table in SSAS.

        Args:
            head (int): The number of records to return from the table.

        Returns:
            list[dict[str, int | float | str]]: A list of SSAS records in dictionary form.
                The keys are the field names and the values are the record values

        """
        return self.tabular_model.server.query_dax(
            f"EVALUATE TOPN({head}, ALL('{self.name}'))",
            db_name=self.tabular_model.db_name,
        )

    def partitions(self) -> set[Partition]:
        """Get associated dependent partitions.

        Returns:
            (set[Partition]): A list of the partitions containing data for this table

        """
        return self.tabular_model.partitions.find_all({"table_id": self.id})

    def columns(self) -> set[Column]:
        """Get associated dependent partitions.

        Returns:
            (set[Column]): A list of the columns in this table

        """
        return self.tabular_model.columns.find_all({"table_id": self.id})

    def default_row_definition(self) -> "DetailRowDefinition | None":
        if self.default_detail_rows_defintion_id is None:
            return None
        return self.tabular_model.detail_row_definitions.find(self.default_detail_rows_defintion_id)

    def table_measures(self) -> set[Measure]:
        """Get measures saved to this table.

        Returns:
            (set[Measure]): A list of measures saved to this table

        Note:
            These measures do not necessarily have calculations that depend on this table.
                For that use `table.measures()`

        """
        return self.tabular_model.measures.find_all({"table_id": self.id})

    def measures(self, *, recursive: bool = False) -> set[Measure]:
        """Get measures that logically depend on this table.

        Examples:
            ```python
            print(measure.expression)
            # sumx(example, [a])

            Table(name=example).measures()
            # [..., Measure(name='measure'), ...]
            ```
        Args:
            recursive (bool): Whether to include measures that depend on other measures.

        Returns:
            (set[Measure]): A list of measures that logically depend this table

        Note:
            These measures are not necessarily saved physically to this table. For that use `table.table_measures()`

        """
        ret = set()
        for col in self.columns():
            ret.update(col.child_measures(recursive=recursive))
        return ret

    def model(self) -> "Model":
        return self.tabular_model.model

    def get_lineage(self, lineage_type: Literal["children", "parents"]) -> LineageNode:
        if lineage_type == "children":
            return LineageNode(
                self,
                lineage_type,
                [col.get_lineage(lineage_type) for col in self.columns()]
                + [partition.get_lineage(lineage_type) for partition in self.partitions()]
                + [measure.get_lineage(lineage_type) for measure in self.measures()],
            )
        return LineageNode(self, lineage_type, [self.model().get_lineage(lineage_type)])

    def refresh(self, *, include_model_refresh: bool = True) -> list[BeautifulSoup]:  # pyright: ignore reportIncompatibleMethodOverride
        """Needs a model refresh to properly propogate the update."""
        if include_model_refresh:
            return [
                super().refresh(),
                self.model().refresh(),
            ]
        return [super().refresh()]
